import datetime
import importlib.util
from importlib.machinery import ModuleSpec
from pathlib import Path
import re

from newssurvey.util.dotenv_ import load_dotenv

load_dotenv()

CWD: Path = Path.cwd()
PACKAGE_PATH: Path = Path(__file__).parent
PACKAGE_NAME: str = PACKAGE_PATH.name

CACHE_EXPIRATION_BY_TAG: dict[str, int] = {tag: datetime.timedelta(weeks=weeks).total_seconds() for tag, weeks in {"get_search_response": 1, "get_article_response": 52, "get_completion": 52, "get_embedding": 52}.items()}
CACHE_SIZES_GiB: dict[str, int] = {"small": 1, "medium": 5, "large": 10}
CITATION_OPEN_CHAR, CITATION_CLOSE_CHAR = "〚〛"  # Defined in the corresponding LLM prompt.
CITATION_GROUP_PATTERN = re.compile(CITATION_OPEN_CHAR + r"(.*?)" + CITATION_CLOSE_CHAR)
GiB = 1024**3
NEWS_SOURCE_NAMESPACE: str = f"{PACKAGE_NAME}.sources"
NEWS_SOURCES: dict[str, ModuleSpec] = {s.name: importlib.util.find_spec(f"{NEWS_SOURCE_NAMESPACE}.{s.name}") for s in (PACKAGE_PATH / "sources").iterdir()}  # Note: A direct import is not practicable here due to a circular import.
NUM_SECTIONS_DEFAULT: int = 100
NUM_SECTIONS_MIN: int = 10  # Applies only to the `max_sections` argument. Does not apply to LLM output.
NUM_SECTIONS_MAX: int = 100
assert NUM_SECTIONS_MIN <= NUM_SECTIONS_DEFAULT <= NUM_SECTIONS_MAX
OUTPUT_FORMAT_DEFAULT: str = "txt"
PROMPTS: dict[str, str] = {p.stem: p.read_text().strip() for p in (PACKAGE_PATH / "prompts").glob("*.txt")}
REQUEST_HEADERS = {"User-Agent": "Mozilla/5.0 (X11; Linux x86_64; rv:129.0) Gecko/20100101 Firefox/129.0"}



import concurrent.futures
from types import ModuleType
from newssurvey.types import SearchArticle, SearchResult


def _get_article(source_module: ModuleType, search_result: SearchResult) -> SearchArticle:
    """Return the article for the given search result."""
    article_text = source_module.get_article_text(search_result["link"])
    article = SearchArticle(**search_result, text=article_text)
    assert article_text.startswith(search_result["title"]), article  # If this fails, fix the parsing to ensure it is true.
    print(f'Received article text for {article["title"]!r}.')
    return article


def get_articles(source_module: ModuleType, search_results: list[SearchResult]) -> list[SearchArticle]:
    """Return a list of articles for the given search results."""
    fetch_article = lambda search_result: _get_article(source_module, search_result)
    with concurrent.futures.ThreadPoolExecutor(max_workers=4) as executor:
        articles = list(executor.map(fetch_article, search_results))
    return articles



import os
import tempfile
from pathlib import Path
from typing import Optional

import click
import datetime
import locket
import pathvalidate

import newssurvey.exceptions
from newssurvey.config import CWD, NUM_SECTIONS_DEFAULT, NUM_SECTIONS_MIN, NUM_SECTIONS_MAX, OUTPUT_FORMAT_DEFAULT, PACKAGE_NAME
from newssurvey.newssurvey import generate_response
from newssurvey.util.openai_ import ensure_openai_key
from newssurvey.util.sys_ import print_error
from newssurvey.workflow.user.query import get_query, ensure_query_is_valid
from newssurvey.workflow.user.source import get_source, get_source_module, ensure_source_is_valid
from newssurvey.workflow.user.output import SUPPORTED_OUTPUT_FORMATS


def _get_default_output_format_and_filename(*, title: str, output_format: Optional[str]) -> str:
    """Return the default output format and filename for the given title."""
    now = datetime.datetime.now().isoformat(timespec="seconds")
    output_stem = f"{now} {title}"
    output_format = output_format or OUTPUT_FORMAT_DEFAULT
    output_name = f"{output_stem}.{output_format}"
    output_name = pathvalidate.sanitize_filename(output_name, platform="auto")
    return output_format, output_name


def _get_output_format_and_path(*, output_format: Optional[str], output_path: Path, title: Optional[str] = None) -> tuple[str, Path]:
    """Return the output format and path, ensuring they are valid."""
    if (output_format is not None) and (output_format not in SUPPORTED_OUTPUT_FORMATS):
        raise newssurvey.exceptions.InputError(f"Output format {output_format!r} is not supported. Supported formats are: {', '.join(SUPPORTED_OUTPUT_FORMATS)}")

    if output_path is None:
        if title:
            output_format, output_filename = _get_default_output_format_and_filename(title=title, output_format=output_format)
            output_path = CWD / output_filename
        else:
            output_format = output_format or OUTPUT_FORMAT_DEFAULT
            output_path = CWD
    else:
        assert isinstance(output_path, Path), (output_path, type(output_path))
    output_path = output_path.expanduser().resolve()

    if output_path.is_dir():
        assert output_path.exists()
        if title:
            output_format, output_filename = _get_default_output_format_and_filename(title=title, output_format=output_format)
            output_path = output_path / output_filename
        else:
            output_format = output_format or OUTPUT_FORMAT_DEFAULT
    else:
        supported_output_suffixes = tuple(f".{suffix}" for suffix in SUPPORTED_OUTPUT_FORMATS)
        if not output_path.name.endswith(supported_output_suffixes):
            raise newssurvey.exceptions.InputError(f"Output file path suffix {output_path.suffix!r} of output file name {output_path.name!r} is not supported. Supported suffixes are: {', '.join(supported_output_suffixes)}. If intended as a directory, it must exist.")
        extracted_output_suffix = next(suffix for suffix in supported_output_suffixes if output_path.name.endswith(suffix))
        extracted_output_format = extracted_output_suffix.lstrip(".")
        assert extracted_output_format in SUPPORTED_OUTPUT_FORMATS, (extracted_output_format, SUPPORTED_OUTPUT_FORMATS)
        if output_format and (output_format != extracted_output_format):
            raise newssurvey.exceptions.InputError(f"Output file path format {extracted_output_format!r} does not match the specified output format {output_format!r}.")
        output_format = extracted_output_format
        output_path.parent.mkdir(parents=True, exist_ok=True)

    pathvalidate.validate_filepath(output_path, platform="auto")
    return output_format, output_path


@click.command(context_settings={"help_option_names": ["-h", "--help"], "max_content_width": 120})
@click.option("--source", "-s", default=None, help="Name of supported news source. If not given, the user is prompted for it.")
@click.option("--query", "-q", default=None, help="Question or concern answerable by the news source. If a path to a file, the file text is read as text. If not given, the user is prompted for it.")
@click.option("--max-sections", "-m", default=NUM_SECTIONS_DEFAULT, type=click.IntRange(NUM_SECTIONS_MIN, NUM_SECTIONS_MAX), help=f"Maximum number of sections to include in the response, between {NUM_SECTIONS_MIN} and {NUM_SECTIONS_MAX}. Its recommended value, also the default, is {NUM_SECTIONS_DEFAULT}.")
@click.option("--output-format", "-f", default=None, help=f"Output format of the response. It can be txt (for text), md (for markdown), gfm.md (for GitHub Flavored markdown), html, pdf, or json. If not specified, but if an output filename is specified via '--output-path', it is determined automatically from the file extension. If not specified, and if an output filename is not specified either, its default is {OUTPUT_FORMAT_DEFAULT}.")
@click.option("--output-path", "-o", default=None, type=Path, help="Output directory path or file path. If intended as a directory path, it must exist, and the file name is auto-determined. If intended as a file path, its extension can be txt (for text), md (for markdown), gfm.md (for GitHub Flavored markdown), html, pdf, or json. If not specified, the output file is written to the current working directory with an auto-determined file name. The response is written to the file except if there is an error.")
@click.option("--confirm/--no-confirm", "-c/-nc", default=True, help="Confirm as the workflow progresses. If `--confirm`, a confirmation is interactively sought as each step of the workflow progresses, and this is the default. If `--no-confirm`, the workflow progresses without any confirmation.")
def main(*args, **kwargs) -> None:
    """Generate and write a response to a question or concern using a supported news source.

    A single instance of this method is enforced.
    """
    lockfile_path = Path(tempfile.gettempdir()) / f"{PACKAGE_NAME}.lock"
    try:
        with locket.lock_file(lockfile_path, timeout=1):
            _main(*args, **kwargs)
    except locket.LockError:
        print_error(f"Only a single instance of {PACKAGE_NAME} can be run at a time. This is intended to avoid remote throttling of requests. Another instance of it may be running using the lock file {lockfile_path}.")
        exit(1)


def _main(source: Optional[str], query: Optional[str], max_sections: int, output_format: str, output_path: Path, confirm: bool) -> None:
    """Generate and write a response to a question or concern using a supported news source.

    The progress is printed to stdout.

    A nonzero exitcode exists if there is an error.
    """
    try:
        ensure_openai_key()

        if not source:
            source = get_source()
        ensure_source_is_valid(source)
        source_module = get_source_module(source)

        query_origin = "arg"
        if not query:
            query = get_query(source_type=source_module.SOURCE_TYPE)
            query_origin = "prompt"
        elif (query_path := Path(query)).is_file():
            assert query_path.exists()
            query = query_path.read_text().strip()
            query_origin = "file"
        ensure_query_is_valid(query)

        assert isinstance(max_sections, int), (max_sections, type(max_sections))
        assert NUM_SECTIONS_MIN <= max_sections <= NUM_SECTIONS_MAX, (max_sections, NUM_SECTIONS_MIN, NUM_SECTIONS_MAX)

        output_format, output_path = _get_output_format_and_path(output_format=output_format, output_path=output_path)
        if (query_origin == "file") and (query_path.resolve() == output_path.resolve()):
            raise newssurvey.exceptions.InputError(f"Output file path {str(output_path.resolve())!r} is the same as the query file path {str(query_path.resolve())!r}.")

        assert isinstance(confirm, bool), (confirm, type(confirm))

        response = generate_response(source=source, query=query, max_sections=max_sections, output_format=output_format, confirm=confirm)
        assert output_format == response.format, (output_format, response.format)

        output_format, output_path = _get_output_format_and_path(output_format=output_format, output_path=output_path, title=response.title)
        if (query_origin == "file") and (query_path.resolve() == output_path.resolve()):
            raise newssurvey.exceptions.InputError(f"Output file path {str(output_path.resolve())!r} is the same as the query file path {str(query_path.resolve())!r}.")

        response_data = response.response
        if isinstance(response_data, str):
            output_path.write_text(response_data)
        elif isinstance(response_data, bytes):
            output_path.write_bytes(response_data)
        else:
            assert False, type(response_data)
        print(f"Wrote response titled {response.title!r} in format {response.format} to {str(output_path.resolve())!r}.")
    except newssurvey.exceptions.Error as exc:
        print_error(str(exc))
        query_sep = "\n" if (isinstance(query, str) and (len(query.splitlines()) > 1)) else " "
        print_error(f"Failed to generate response for source {source} for query:{query_sep}{query}")
        exit(1)
    except KeyboardInterrupt:
        print()  # This separates "^C" from the subsequent error.
        print_error("Interrupted by user.")
        os._exit(-2)  # Plain `exit` is not used because it may not immediately terminate, with any background threads potentially still running.


if __name__ == "__main__":
    main()



from typing import Literal

import numpy as np
from scipy.spatial.distance import cosine, euclidean, seuclidean, sqeuclidean

from newssurvey.util.openai_ import EmbeddingModelSizeType, get_vector, get_vectors_concurrently

DistanceType = Literal["cosine", "euclidean", "sqeuclidean", "seuclidean"]


def sort_by_distance(reference: str, items: list[str], *, model_size: EmbeddingModelSizeType, distance: str = DistanceType) -> list[str]:
    """Return items sorted by distance to the reference using the distance function."""
    reference_vector = get_vector(reference, model_size=model_size)
    item_vectors = get_vectors_concurrently(items, model_size=model_size)

    match distance:
        case "cosine" | "euclidean" | "sqeuclidean":
            fn_distance = {"cosine": cosine, "euclidean": euclidean, "sqeuclidean": sqeuclidean}[distance]
            distances = {item: fn_distance(reference_vector, item_vector) for item, item_vector in item_vectors.items()}
        case "seuclidean":
            variances = np.var(list(item_vectors.values()), axis=0)
            distances = {item: seuclidean(reference_vector, item_vector, variances) for item, item_vector in item_vectors.items()}
        case _:
            assert False, distance

    items = sorted(items, key=distances.get)  # Note: The sort is intentionally not done in-place to avoid modifying the input list.
    # print(f"ITEM DISTANCES: ({distance})\n" + "\n".join([f"{num}. {item}: {distances[item]:.4f}" for num, item in enumerate(items, start=1)]))

    return items



import datetime
import json
import re

from newssurvey.config import CITATION_GROUP_PATTERN, PROMPTS
from newssurvey.types import SectionGen2, CitationGen2

_DISCLAIMER = PROMPTS["disclaimer"]
SUPPORTED_OUTPUT_FORMATS: list[str] = ["txt", "gfm.md", "md", "html", "pdf", "json"]  # Note: 'gfm.md' must remain before 'md' to avoid matching 'md' first.


def _get_date_string() -> str:
    """Return the current date as a string."""
    date = datetime.date.today()
    return f"{date.strftime('%B')} {date.day}, {date.year}"  # Platform-independent date formatting. Example: "September 8, 2024"


def format_text_output(title: str, sections: list[SectionGen2], citations: list[CitationGen2]) -> str:
    """Return the text string output for the given sections and citations."""
    sections = [SectionGen2(title=section["title"], text=CITATION_GROUP_PATTERN.sub(r"[\1]", section["text"])) for section in sections]  # Uses standard brackets for citations.

    text = f"{title}\n\n" + f"{_get_date_string()}\n\n{_DISCLAIMER}\n\n" + "Sections:\n" + "\n".join([f"{num}: {section['title']}" for num, section in enumerate(sections, start=1)]) + "\n\n" + "\n\n".join(f'Section {num}: {s["title"]}:\n\n{s["text"]}' for num, s in enumerate(sections, start=1)) + "\n\n" + "References:\n\n" + "\n\n".join([f"{c['number']}: {c['title']}\n{c['link']}" for c in citations])
    return text


def format_markdown_output(title: str, sections: list[SectionGen2], citations: list[CitationGen2]) -> str:
    """Return the markdown string output for the given sections and citations."""
    contents = [f"{num}. [{section['title']}](#section-{num})" for num, section in enumerate(sections, start=1)]
    contents.append(f"{len(contents) + 1}. [References](#references)")

    def repl(match: re.Match) -> str:
        """Return the match text with the plain citation numbers in the citation group replaced with linked citation numbers."""
        return "<sup>[" + ",".join(f"[{citation_num}](#citation-{citation_num})" for citation_num in match.group(1).split(",")) + "]</sup>"

    sections = [SectionGen2(title=section["title"], text=CITATION_GROUP_PATTERN.sub(repl, section["text"])) for section in sections]

    text = f"# {title}\n\n" + f"{_get_date_string()}\n\n_{_DISCLAIMER}_\n\n" + "## Contents\n" + "\n".join(contents) + "\n\n" + "\n\n".join(f'## <a id="section-{num}"></a>{num}. {s["title"]}\n\n{s["text"]}' for num, s in enumerate(sections, start=1)) + "\n\n" + '## <a id="references"></a>References\n\n' + "\n\n".join([f'<a id="citation-{c["number"]}"></a>{c["number"]}. [{c["title"]}]({c["link"]})' for c in citations])
    return text


def format_gfm_output(title: str, sections: list[SectionGen2], citations: list[CitationGen2]) -> str:
    """Return the GitHub Flavored markdown string output for the given sections and citations."""
    contents = [f"{num}. [{section['title']}](#section-{num})" for num, section in enumerate(sections, start=1)]
    contents.append(f"{len(contents) + 1}. [References](#references)")

    def repl(match: re.Match) -> str:
        """Return the match text with the plain citation numbers in the citation group replaced with GitHub Flavored citation numbers."""
        return "".join(f"[^{citation_num}]" for citation_num in match.group(1).split(","))

    sections = [SectionGen2(title=section["title"], text=CITATION_GROUP_PATTERN.sub(repl, section["text"])) for section in sections]

    text = f"# {title}\n\n" + f"{_get_date_string()}\n\n_{_DISCLAIMER}_\n\n" + "## Contents\n" + "\n".join(contents) + "\n\n" + "\n\n".join(f'## <a id="section-{num}"></a>{num}. {s["title"]}\n\n{s["text"]}' for num, s in enumerate(sections, start=1)) + "\n\n" + '## <a id="references"></a>References\n\n' + "\n\n".join([f'[^{c["number"]}]: [{c["title"]}]({c["link"]})' for c in citations])
    return text


def format_html_output(title: str, sections: list[SectionGen2], citations: list[CitationGen2]) -> str:
    """Return the HTML string output for the given sections and citations."""
    citation_map = {str(citation["number"]): citation for citation in citations}

    contents = [f'<li><a href="#section-{num}">{section["title"]}</a></li>' for num, section in enumerate(sections, start=1)]
    contents.append('<li><a href="#references">References</a></li>')

    def repl(match: re.Match) -> str:
        """Return the match text with the plain citation numbers in the citation group replaced with linked citation numbers and separate hover tooltips."""
        return "<sup>[</sup>" + "<sup>,</sup>".join(f'<a href="#citation-{citation_num}" class="citation-link"><sup>{citation_num}</sup></a>' f'<a href="{citation_map[citation_num]["link"]}" class="citation-tooltip" target="_blank">{citation_map[citation_num]["title"]}</a>' for citation_num in match.group(1).split(",")) + "<sup>]</sup>"

    def format_section_text(text: str) -> str:
        """Return the section text wrapped in HTML paragraph tags and replace citation numbers with linked citation numbers."""
        paragraphs = text.split("\n\n")
        wrapped_paragraphs = [f"<p>{CITATION_GROUP_PATTERN.sub(repl, paragraph.strip())}</p>" for paragraph in paragraphs]
        return "\n".join(wrapped_paragraphs)

    sections_html = [f'<h2 id="section-{num}">{num}. {section["title"]}</h2>\n{format_section_text(section["text"])}' for num, section in enumerate(sections, start=1)]

    references_html = [f'<li id="citation-{citation["number"]}"><a href="{citation["link"]}" target="_blank">{citation["title"]}</a></li>' for citation in citations]

    html_output = f"""
<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>{title}</title>
    <link href="https://fonts.googleapis.com/css2?family=Roboto:wght@400;700&family=Merriweather:wght@400;700&display=swap" rel="stylesheet">
    <style>
        body {{
            font-family: 'Roboto', sans-serif;
            margin: 40px;
            line-height: 1.6;
        }}
        h1, h2 {{
            font-family: 'Merriweather', serif;
            color: #333;
        }}
        h1 {{
            font-size: 2.5em;
            margin-bottom: 0.5em;
        }}
        h2 {{
            font-size: 1.75em;
            margin-top: 2em;
        }}
        p {{
            font-size: 1.1em;
        }}
        ol {{
            padding-left: 20px;
        }}
        a {{
            color: #1a73e8;
            text-decoration: none;
        }}
        a:hover {{
            text-decoration: underline;
        }}

        .citation-link {{
            color: #1a73e8;
            text-decoration: none;
        }}
        .citation-link:hover + .citation-tooltip {{
            display: inline-block;
        }}
        .citation-tooltip {{
            display: none;
            position: absolute;
            background-color: #f9f9f9;
            border: 1px solid #ccc;
            padding: 5px;
            font-size: 0.9em;
            white-space: nowrap;
            z-index: 10;
            box-shadow: 0px 4px 8px rgba(0, 0, 0, 0.1);
            cursor: pointer;
            color: #1a73e8;
        }}
        .citation-tooltip:hover {{
            display: inline-block;
        }}

        /* Mobile responsiveness */
        @media (max-width: 600px) {{
            body {{
                margin: 20px;
            }}
            h1 {{
                font-size: 2em;
            }}
            h2 {{
                font-size: 1.5em;
            }}
            ol {{
                padding-left: 15px;
            }}
        }}
    </style>
</head>
<body>

<h1>{title}</h1>
<p>{_get_date_string()}</p>
<p><em>{_DISCLAIMER}</em></p>

<h2>Contents</h2>
<ol>
    {'\n    '.join(contents)}
</ol>

{'\n\n'.join(sections_html)}

<h2 id="references">References</h2>
<ol>
    {'\n    '.join(references_html)}
</ol>

</body>
</html>
    """

    return html_output


def format_json_output(title: str, sections: list[SectionGen2], citations: list[CitationGen2]) -> str:
    """Return a JSON string output for the given sections and citations."""
    data = {
        "title": title,
        "date": _get_date_string(),
        "disclaimer": _DISCLAIMER,
        "sections": [{"number": num, "title": section["title"], "text": section["text"]} for num, section in enumerate(sections, start=1)],
        "citations": [{"number": citation["number"], "title": citation["title"], "link": citation["link"]} for citation in citations],
    }
    return json.dumps(data, indent=4)


def format_pdf_output(title: str, sections: list[SectionGen2], citations: list[CitationGen2]) -> bytes:
    """Return the PDF bytes output for the given sections and citations."""
    from io import BytesIO
    from reportlab.platypus import Flowable, SimpleDocTemplate, Paragraph, Spacer, PageBreak, ListFlowable, ListItem
    from reportlab.lib.styles import getSampleStyleSheet, ParagraphStyle
    from reportlab.lib.pagesizes import letter
    from reportlab.lib.units import inch
    from reportlab.pdfgen.canvas import Canvas

    class MyDocTemplate(SimpleDocTemplate):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._bookmark_id = 0

        def afterFlowable(self, flowable: Flowable) -> None:
            """Add bookmark and outline entries for the flowable if a Paragraph with a style of "Title", "Heading1", or "Heading2"."""
            if isinstance(flowable, Paragraph):
                text = flowable.getPlainText()
                style_name = flowable.style.name
                level = 0 if style_name == "Title" else 1 if style_name in ("Heading1", "Heading2") else None
                if level is not None:
                    key = f"bk_{self._bookmark_id}"
                    self._bookmark_id += 1
                    self.canv.bookmarkPage(key, fit="XYZ", left=0, zoom=None)
                    self.canv.addOutlineEntry(text, key, level=level, closed=False)

        def onPage(self, canvas: Canvas, doc: SimpleDocTemplate) -> None:
            """Add the page number to the footer of each page."""
            page_num = doc.page
            text = f"{page_num}"
            canvas.saveState()
            canvas.setFont("Helvetica", 9)
            canvas.drawCentredString(letter[0] / 2.0, 0.5 * inch, text)  # Page number at bottom center
            canvas.restoreState()

    buffer = BytesIO()
    doc = MyDocTemplate(buffer, pagesize=letter, rightMargin=72, leftMargin=72, topMargin=72, bottomMargin=72)
    styles = getSampleStyleSheet()
    styles.add(ParagraphStyle(name="TOC", parent=styles["Normal"], fontSize=12, leading=14, leftIndent=20))
    styles.add(ParagraphStyle(name="Center", parent=styles["Normal"], alignment=1))

    story = [Paragraph(title, styles["Title"]), Spacer(1, 12), Paragraph(_get_date_string(), styles["Center"]), Spacer(1, 12), Paragraph(_DISCLAIMER, styles["Italic"]), Spacer(1, 12), Paragraph('<a name="contents"/>Contents', styles["Heading1"]), Spacer(1, 12)]

    toc_items = [ListItem(Paragraph(f'<a href="#section_{num}">{section["title"]}</a>', styles["TOC"])) for num, section in enumerate(sections, start=1)]
    toc_items.append(ListItem(Paragraph('<a href="#references">References</a>', styles["TOC"])))
    story += [ListFlowable(toc_items, bulletType="1"), PageBreak()]

    def replace_citations(text):
        return CITATION_GROUP_PATTERN.sub(lambda match: "<super>" + ",".join([f'<a href="#citation_{num.strip()}">{num.strip()}</a>' for num in match.group(1).split(",")]) + "</super>", text)

    num_sections = len(sections)
    for section_num, section in enumerate(sections, start=1):
        story += [Paragraph(f'{section_num}. <a name="section_{section_num}"/>{section["title"]}', styles["Heading1"]), Spacer(1, 12)]
        paragraphs = replace_citations(section["text"]).split("\n\n")
        num_paragraphs = len(paragraphs)
        for para_num, para_text in enumerate(paragraphs, start=1):
            story.append(Paragraph(para_text, styles["Normal"]))
            if (section_num != num_sections) or (para_num != num_paragraphs):
                story.append(Spacer(1, 12))

    story += [PageBreak(), Paragraph('<a name="references"/>References', styles["Heading1"]), Spacer(1, 12)]

    num_citations = len(citations)
    for citation_num, citation in enumerate(citations, start=1):
        citation_text = f'<a name="citation_{citation["number"]}"/><b>{citation["number"]}.</b> <a href="{citation["link"]}">{citation["title"]}</a><br/><a href="{citation["link"]}">{citation["link"]}</a>'
        story.append(Paragraph(citation_text, styles["Normal"]))
        if citation_num != num_citations:
            story.append(Spacer(1, 12))

    doc.build(story, onFirstPage=doc.onPage, onLaterPages=doc.onPage)
    pdf_value = buffer.getvalue()
    buffer.close()
    return pdf_value


def format_output(*, title: str, sections: list[SectionGen2], citations: list[CitationGen2], output_format: str, **kwargs) -> str:
    """Return the formatted output for the given sections and citations in the specified format."""
    if output_format not in SUPPORTED_OUTPUT_FORMATS:
        raise ValueError(f"Unsupported output format: {output_format!r}")

    formatters = {
        "txt": format_text_output,
        "md": format_markdown_output,
        "gfm.md": format_gfm_output,
        "html": format_html_output,
        "pdf": format_pdf_output,
        "json": format_json_output,
    }
    formatter = formatters[output_format]
    output = formatter(title=title, sections=sections, citations=citations, **kwargs)
    return output



import itertools
import re

import hext
import requests

from newssurvey.config import CACHE_EXPIRATION_BY_TAG, CACHE_SIZES_GiB
from newssurvey.util.diskcache_ import get_diskcache
from newssurvey.util.sys_ import print_error

from ._common import REQUEST_HEADERS, request_cooldown_lock

_DISKCACHE = get_diskcache(__file__, size_gib=CACHE_SIZES_GiB["large"])
_HEXT = hext.Rule("""
    <html>
        <body>
            { <*:type-matches(/^(p)$|^(h[1-6])$/) @text:content /> }
        </body>
    </html>
    """)  # Ref: https://github.com/html-extract/hext/issues/30#issuecomment-2072869699
_PRE_CONTENT_BLACKLIST = {
    "Science X Account",
    "Forget Password?",
    "Learn more",
    "share this!",
    "Twit",
    "Share",
    "Email",
}
_PRE_CONTENT_STARTSWITH_BLACKLIST = ("Click here to ",)
_PRE_CONTENT_RE_FULLMATCH_BLACKLIST = [
    re.compile(p)
    for p in (
        r"\d+(?:\.\d+)?K",  # Example: 1.9K
        r"\d+",  # Example: 39
    )
]
_MID_CONTENT_BLACKLIST = {
    "written by researcher(s)",
    #
    # Review attributes:
    "fact-checked",
    "peer-reviewed publication",
    "trusted source",
    "proofread",
    "reputable news agency",
}
_MID_CONTENT_STARTSWITH_BLACKLIST = (
    "This article has been reviewed ",
    "©",
    "Copyright ©",
)
_POST_CONTENT_BLACKLIST = {
    "Explore further",
}
_POST_CONTENT_STARTSWITH_BLACKLIST = ()
_CONTENT_SUFFIX_REMOVELIST = (" Read the original article.",)


@_DISKCACHE.memoize(expire=CACHE_EXPIRATION_BY_TAG["get_article_response"], tag="get_article_response")
def _get_article_response(url: str) -> requests.Response:
    with request_cooldown_lock:
        print(f"Reading {url}.")
        response = requests.get(url, headers=REQUEST_HEADERS)
        # Note: params={"deviceType": "mobile"} is not specified because it results in the loss of the article date.
    try:
        response.raise_for_status()
    except requests.RequestException:
        print_error(f"Failed to read {url} due to status code {response.status_code}.")
        raise
    print(f"Read {url} with status code {response.status_code}.")
    return response


def _get_article_content(url: str) -> list[str]:
    response = _get_article_response(url)
    html = response.text

    rule = _HEXT
    html = hext.Html(html)
    results = rule.extract(html)
    results = results[0]["content"]

    results = [r.strip() for r in results]
    results = [r for r in results if r]

    return results


def _get_filtered_article_content(url: str) -> list[str]:
    content = _get_article_content(url)
    assert content

    # Remove invalid pre-article content
    dropwhile_predicate = lambda c: ((c in _PRE_CONTENT_BLACKLIST) or (c.startswith(_PRE_CONTENT_STARTSWITH_BLACKLIST)) or any(p.fullmatch(c) for p in _PRE_CONTENT_RE_FULLMATCH_BLACKLIST))
    content = list(itertools.dropwhile(dropwhile_predicate, content))

    # Remove invalid post-article content
    takewhile_predicate = lambda c: ((c not in _POST_CONTENT_BLACKLIST) and (not c.startswith(_POST_CONTENT_STARTSWITH_BLACKLIST)))
    content = list(itertools.takewhile(takewhile_predicate, content))

    # Remove invalid mid-article content
    content = [c for c in content if ((c not in _MID_CONTENT_BLACKLIST) and (not c.startswith(_MID_CONTENT_STARTSWITH_BLACKLIST)))]

    # Remove invalid suffix
    for idx, c in enumerate(content.copy()):
        for suffix in _CONTENT_SUFFIX_REMOVELIST:
            if c.endswith(suffix):
                content[idx] = c.removesuffix(suffix)

    assert content
    return content


def get_article_text(url: str) -> str:
    content = _get_filtered_article_content(url)
    text = "\n\n".join(content)
    text = text.strip()
    return text



import threading
import time
from typing import Optional, Self


class CooldownLock:
    """A lock mechanism that also enforces a cooldown period between releases and subsequent acquisitions.

    This is useful in scenarios where it's also necessary to limit the frequency of lock acquisition,
    such as for managing access to a shared resource in a way that prevents excessively rapid consecutive uses.

    Ref: https://stackoverflow.com/a/78591650/
    """

    def __init__(self, cooldown: float = 1, name: Optional[str] = None):
        """Initialize with a specified cooldown period.

        Args:
            cooldown: The cooldown period in seconds after releasing the lock before it can be acquired again.
            name: Optional name of lock used in log messages.
        """
        self._cooldown_period = cooldown
        self._name = f"{name} lock" if name else "lock"

        self._earliest_use_time = 0
        self._lock = threading.Lock()

    def acquire(self) -> bool:
        self._lock.acquire()
        wait_time = self._earliest_use_time - time.monotonic()
        if wait_time > 0:
            print(f"Sleeping for {wait_time:.1f}s to acquire {self._name}.")
            time.sleep(wait_time)
        return True

    def release(self):
        if self.locked():
            self._earliest_use_time = time.monotonic() + self._cooldown_period
        self._lock.release()

    def locked(self) -> bool:
        return self._lock.locked()

    def __enter__(self) -> Self:
        self.acquire()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.release()



from functools import cache
from pathlib import Path

import diskcache

from newssurvey.config import CACHE_SIZES_GiB, GiB, PACKAGE_PATH


DISKCACHE_ROOT_PATH = PACKAGE_PATH / ".diskcache"


@cache
def get_diskcache(file_path: str, *, size_gib: int = CACHE_SIZES_GiB["small"]) -> diskcache.FanoutCache:
    """Return a diskcache object for the given file path.

    `file_path` should typically be `__file__`.
    """
    path = Path(file_path)  # Ex: /workspaces/newssurvey/src/newssurvey/sources/medicalxpress/article.py
    assert path.is_relative_to(PACKAGE_PATH), path
    path = path.relative_to(PACKAGE_PATH)  # Ex: sources/medicalxpress/article.py
    path = path.with_suffix("")  # Ex: sources/medicalxpress/article
    path = DISKCACHE_ROOT_PATH / path  # Ex: /workspaces/newssurvey/src/newssurvey/.diskcache/sources/medicalxpress/article
    print(f"Using diskcache path: {path}")
    return diskcache.FanoutCache(directory=str(path), timeout=10, size_limit=size_gib * GiB)



import contextlib
import io
from types import ModuleType

from newssurvey.config import PROMPTS
from newssurvey.exceptions import LanguageModelOutputStructureError
from newssurvey.util.openai_ import get_content
from newssurvey.util.sys_ import print_error, print_warning


def _is_title_valid(title: str) -> bool:
    """Return true if the title is valid, otherwise false.

    A validation error is printed if the title is invalid.
    """
    if not title:
        print_error("The title is empty.")
        return False

    if title != title.strip():
        print_error(f"The title has leading or trailing whitespace: {title!r}")
        return False

    num_lines = title.splitlines()
    if len(num_lines) != 1:
        print_error(f"The title has {num_lines} lines but a single line was expected: {title!r}")
        return False

    return True


def create_title(user_query: str, source_module: ModuleType, *, sections: list[str], max_attempts: int = 1) -> str:
    """Return a title using the given section names.

    `LanguageModelOutputError` is raised if the model output has an error.
    Specifically, its subclass `LanguageModelOutputStructureError` is raised by it if the output is structurally invalid.
    """
    assert sections

    numbered_sections = [f"{i}. {s}" for i, s in enumerate(sections, start=1)]
    numbered_sections_str = "\n".join(numbered_sections)

    prompt_data = {"user_query": user_query, "source_site_name": source_module.SOURCE_SITE_NAME, "source_type": source_module.SOURCE_TYPE}
    prompt_data["task"] = PROMPTS["7. create_title"].format(sections=numbered_sections_str)
    prompt = PROMPTS["0. common"].format(**prompt_data)

    for num_attempt in range(1, max_attempts + 1):
        response = get_content(prompt, model_size="large", log=True, read_cache=(num_attempt == 1))

        error = io.StringIO()
        with contextlib.redirect_stderr(error):
            response_is_valid = _is_title_valid(response)
        if not response_is_valid:
            error = error.getvalue().rstrip().removeprefix("Error: ")
            if num_attempt == max_attempts:
                raise LanguageModelOutputStructureError(error)
            else:
                print_warning(f"Fault in attempt {num_attempt} of {max_attempts} while getting title: {error}")
                continue

        break

    return response



class Error(Exception):
    """Base error."""


class EnvError(Error, EnvironmentError):
    """Environment variable availability error."""


class InputError(Error, ValueError):
    """Input validity error."""


class ModelOutputError(Error):
    """Model output error."""


class LanguageModelOutputError(ModelOutputError):
    """Language model output error."""


class LanguageModelOutputConvergenceError(LanguageModelOutputError):
    """Language model output convergence error."""


class LanguageModelOutputLimitError(LanguageModelOutputError):
    """Language model output limit error."""


class LanguageModelOutputRejectionError(LanguageModelOutputError):
    """Language model output rejection error."""


class LanguageModelOutputStructureError(LanguageModelOutputError):
    """Language model output structure error."""


class RequestError(Error):
    """Malformed HTTP request error"""


class SourceInsufficiencyError(Error):
    """Source data insufficiency error."""



import contextlib
import importlib
import io
from types import ModuleType

import newssurvey.exceptions
from newssurvey.config import NEWS_SOURCES, NEWS_SOURCE_NAMESPACE
from newssurvey.util.sys_ import print_error


def is_source_valid(source: str) -> bool:
    """Return true if the news source is supported, otherwise false.

    A validation error is printed if the news source is unsupported.
    """
    if not isinstance(source, str):
        print_error("Source must be a string.")
        return False
    if source != source.strip():
        print_error("Source must not have leading or trailing whitespace.")  # It is assumed that whitespace is stripped before this check.
        return False
    if len(source) == 0:
        print_error("No source was provided.")
        return False
    if source not in NEWS_SOURCES:
        supported_news_sources = ", ".join(sorted(NEWS_SOURCES))
        print_error(f"{source!r} is not among the supported news sources: {supported_news_sources}")
        return False
    if not NEWS_SOURCES[source]:
        print_error(f"{source!r} is not an importable news source module.")
        return False
    return True


def ensure_source_is_valid(source: str) -> None:
    """Raise `InputError` if the news source is unsupported."""
    error = io.StringIO()
    with contextlib.redirect_stderr(error):
        if not is_source_valid(source):
            error = error.getvalue().rstrip().removeprefix("Error: ")
            raise newssurvey.exceptions.InputError(error)


def get_source() -> str:
    """Get news source from user input."""
    assert NEWS_SOURCES
    supported_news_sources = dict(zip(map(str, range(1, len(NEWS_SOURCES) + 1)), sorted(NEWS_SOURCES)))
    print("The supported news sources are:")
    for source_num, source in supported_news_sources.items():
        print(f"\t{source_num}. {source}")

    source = None
    while not source:
        source = input("Specify the name or number of a supported news source: ")
        source = source.strip().lower()
        if source in supported_news_sources:
            source = supported_news_sources[source]

        if not is_source_valid(source):
            source = None
    return source


def get_source_module(source: str) -> ModuleType:
    """Get the source module corresponding to the given source name."""
    return importlib.import_module(NEWS_SOURCES[source].name)


def get_source_module_name(module: ModuleType) -> str:
    name = module.__name__
    prefix = f"{NEWS_SOURCE_NAMESPACE}."
    assert name.startswith(prefix), name
    name = name.removeprefix(prefix)
    assert "." not in name, name
    return name



from typing import Callable

import hext
import requests

from newssurvey.config import CACHE_EXPIRATION_BY_TAG, CACHE_SIZES_GiB
from newssurvey.exceptions import RequestError
from newssurvey.types import SearchResult
from newssurvey.util.diskcache_ import get_diskcache
from newssurvey.util.sys_ import print_error

from ._common import REQUEST_HEADERS, request_cooldown_lock

_DISKCACHE = get_diskcache(__file__, size_gib=CACHE_SIZES_GiB["small"])
_HEXT = hext.Rule("""
    <h2 class="mb-2">
      <a href:link class="news-link" @text:title />
    </h2>
    <p class="mb-4" @text:description />
    """)
MAX_PAGE_NUM = 40  # Observed empirically. Ex: https://medicalxpress.com/search/page41.html?search=papaya&s=0


class UnsupportedPageError(RequestError):
    """Excessive page number request error."""


@_DISKCACHE.memoize(expire=CACHE_EXPIRATION_BY_TAG["get_search_response"], tag="get_search_response")
def _get_search_response(query: str, *, sort_by: str = "relevancy", headlines: bool = False, page_num: int = 1) -> requests.Response:
    """Return a response from the MedicalXpress website for a given query, sorting preference, and page number.

    Parameters:
        query (str): The search term used to query the website.
        sort_by (str, optional): The sorting method for the search results ('relevancy' or 'date'), with a default value of 'relevancy'.
        headlines (bool): The filter to limit search results to headlines matches only, with a default value of False.
        page_num (int, optional): The page number to retrieve, with a default value of 1.

    Returns:
        requests.Response: The HTTP response object containing the search results.

    Raises:
        UnsupportedPageError: If the page number is higher than the maximum allowed.
    """
    url = f"https://medicalxpress.com/search/page{page_num}.html"
    params = {"search": query, "s": {"relevancy": 0, "date": 1}[sort_by], "h": {True: 1, False: 0}[headlines]}
    headlines_filter_status = "with" if headlines else "without"
    description = f'page {page_num} of search results {headlines_filter_status} the headlines filter for "{query}" sorted by {sort_by}'
    if page_num > MAX_PAGE_NUM:
        raise UnsupportedPageError(f"Unable to request {description} because it exceeds the max page number of {MAX_PAGE_NUM}.")
    with request_cooldown_lock:
        print(f"Requesting {description}.")
        response = requests.get(url, params=params, headers=REQUEST_HEADERS)
    try:
        response.raise_for_status()
    except requests.RequestException:
        print_error(f"Failed to receive {description} due to status code {response.status_code}.")
        raise
    print(f"Received {description} with status code {response.status_code}.")
    return response


def get_search_results(**kwargs) -> list[SearchResult]:
    """Return search results as a list of dictionaries, each containing the 'title', 'link', and 'description' of an article.

    `kwargs` are forwarded to `_get_search_response`.

    Returns:
        list[dict]: A list of dictionaries where each dictionary represents an article with its title, link, and description.

    An empty list is returned if the requested page number exceeds its maximum limit.
    """
    try:
        response = _get_search_response(**kwargs)
    except UnsupportedPageError:
        return []
    html = response.text
    html = hext.Html(html)
    rule = _HEXT
    results = rule.extract(html)

    for result in results:
        result["title"] = result["title"].rstrip()  # Removes observed trailing character like \xa0, \u202f, etc.
        assert result["title"] == result["title"].strip(), (result["title"], result["title"].strip())  # If leading whitespace is observed, it may then have to be removed from both the title and the text.

    return results


def run_searches(target: Callable) -> None:
    """Call the given target function with the respective keyword arguments for each type of search that can be conducted.

    Note: The keyword arguments `query` and `page_num` are managed automatically by the target function, and are therefore not provided.
    """
    for sort_by in ("relevancy", "date"):
        for headlines in (True, False):
            target(sort_by=sort_by, headlines=headlines)



import concurrent.futures
import contextlib
import io
import textwrap
from types import ModuleType
from typing import Optional

from newssurvey.config import PROMPTS
from newssurvey.exceptions import LanguageModelOutputStructureError, SourceInsufficiencyError
from newssurvey.types import SearchArticle, AnalyzedArticleGen1, AnalyzedArticleGen2, AnalyzedSectionGen1, AnalyzedSectionGen2
from newssurvey.util.openai_ import get_content, MAX_WORKERS
from newssurvey.util.str import is_none_response
from newssurvey.util.sys_ import print_warning, print_error


def _is_output_valid(text: str, article_title: str, section_name: str) -> bool:
    """Return true if the output text is valid, otherwise false.

    A validation error is printed if the output text is invalid.
    """
    if text != text.strip():
        print_error(f"The text for the article {article_title!r} for the section {section_name!r} has leading or trailing whitespace.")
        return False

    if not text:
        print_error(f"The text for the article {article_title!r} for the section {section_name!r} is empty.")
        return False

    def is_valid_starting_line(starting_line: str, invalid_line: str) -> bool:
        invalid_starting_lines = (invalid_line, f"{invalid_line}:", f"**{invalid_line}**", f"**{invalid_line}**:")
        return starting_line not in invalid_starting_lines

    starting_line = text.split("\n", 1)[0]
    for invalid_line in (article_title, section_name):
        if not is_valid_starting_line(starting_line, invalid_line):
            print_error(f"The text for the article {article_title!r} for the section {section_name!r} starts with the invalid line: {starting_line!r}")
            return False

    return True


def _condense_article(user_query: str, source_module: ModuleType, *, article: SearchArticle, sections: list[str], section: str, max_attempts: int = 3) -> Optional[str]:
    assert user_query
    assert sections

    numbered_sections = [f"{i}. {s}" for i, s in enumerate(sections, start=1)]
    numbered_sections_str = "\n".join(numbered_sections)
    numbered_section = f"{sections.index(section) + 1}. {section}"

    prompt_data = {"user_query": user_query, "source_site_name": source_module.SOURCE_SITE_NAME, "source_type": source_module.SOURCE_TYPE}
    assert article["text"].startswith(article["title"]), article  # If this fails, fix the parsing to ensure it is true.

    prompt_data["task"] = PROMPTS["5. condense_article"].format(num_sections=len(sections), sections=numbered_sections_str, section=numbered_section, article=article["text"])
    prompt = PROMPTS["0. common"].format(**prompt_data)

    for num_attempt in range(1, max_attempts + 1):
        response = get_content(prompt, model_size="small", log=(num_attempt > 1), read_cache=(num_attempt == 1))

        if is_none_response(response):
            return

        error = io.StringIO()
        with contextlib.redirect_stderr(error):
            response_is_valid = _is_output_valid(response, article["title"], section)
        if not response_is_valid:
            error = error.getvalue().rstrip().removeprefix("Error: ")
            if num_attempt == max_attempts:
                raise LanguageModelOutputStructureError(error)
            else:
                print_warning(f"Fault in attempt {num_attempt} of {max_attempts} while getting a condensed article: {error}")
                continue

        break

    return response


def condense_articles(user_query: str, source_module: ModuleType, *, articles: list[AnalyzedArticleGen1], sections: list[str]) -> list[AnalyzedArticleGen2]:
    """Return a list of dictionaries containing the given search article, the given rated section names, with the corresponding condensed text for each section.

    The returned text for each article-section pair is the condensed version of the article in the context of the section and the user query.

    Article-section pairs with no text are skipped. Articles with no remaining sections are also skipped.

    The internal functions raise `LanguageModelOutputError` if the model output has an error.
    Specifically, its subclass `LanguageModelOutputStructureError` is raised by it if the output is structurally invalid.

    `SourceInsufficiencyError` is raised if no usable articles remain for the query.
    """
    assert articles
    assert sections
    # articles = sorted(articles, key=lambda a: len(a["article"]["link"]), reverse=True)  # For reproducible testing, but not necessary for cache.

    def condense_article_section(article: SearchArticle, section: AnalyzedSectionGen1) -> Optional[str]:
        assert section["rating"] > 0
        return _condense_article(user_query, source_module, article=article, sections=sections, section=section["section"])

    condensed_articles = []

    max_workers = min(8, MAX_WORKERS)  # Note: Using a value ≥10 caused diskcache to hang indefinitely.
    with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
        future_to_section = {executor.submit(condense_article_section, article["article"], section): (article, section) for article in articles for section in article["sections"]}

        article_section_results = {}
        for future in concurrent.futures.as_completed(future_to_section):
            article, section = future_to_section[future]
            article_title = article["article"]["title"]
            section_name = section["section"]
            condensed_text = future.result()
            if condensed_text is None:
                print(f"There is no condensed text for the article {article_title!r} for the section {section_name!r}.")
                continue
            print(f'The condensed text for the article {article_title!r} for the section {section_name!r} with rating {section['rating']} is:\n{textwrap.indent(condensed_text, prefix="\t")}')
            article_section_results[(article_title, section_name)] = condensed_text

    for article in articles:
        article_title = article["article"]["title"]
        condensed_sections = []
        for section in article["sections"]:
            section_name = section["section"]
            result_key = (article_title, section_name)
            if result_key in article_section_results:
                condensed_text = article_section_results[result_key]
                condensed_section = AnalyzedSectionGen2(**section, text=condensed_text)
                condensed_sections.append(condensed_section)
        if not condensed_sections:
            num_article_sections = len(article["sections"])
            print_warning(f"There is no condensed text for any of the {num_article_sections} sections for the article {article_title!r}.")
            continue
        condensed_articles.append(AnalyzedArticleGen2(article=article["article"], sections=condensed_sections))

    condensed_articles.sort(key=lambda a: sum(s["rating"] for s in a["sections"]), reverse=True)
    if not condensed_articles:
        raise SourceInsufficiencyError("No usable articles exist for the query.")
    return condensed_articles



from newssurvey.config import REQUEST_HEADERS
from newssurvey.util.threading_ import CooldownLock

_REQUEST_COOLDOWN = 3  # Applicable to default Mozilla Firefox user agent.
REQUEST_HEADERS = REQUEST_HEADERS.copy()
# SOURCE_SITE_URL = "medicalxpress.com"
SOURCE_SITE_NAME = "MedicalXpress"
SOURCE_TYPE = "medical"

request_cooldown_lock = CooldownLock(cooldown=_REQUEST_COOLDOWN, name=SOURCE_SITE_NAME)  # Note: 2s or lower values of cooldown led to HTTP 429 error.



import contextlib
import io
import re
from types import ModuleType
from typing import Optional

from newssurvey.config import PROMPTS, NUM_SECTIONS_MIN, NUM_SECTIONS_MAX
from newssurvey.exceptions import LanguageModelOutputRejectionError, LanguageModelOutputStructureError
from newssurvey.util.openai_ import get_content, MODELS
from newssurvey.util.scipy_ import sort_by_distance
from newssurvey.util.str import is_none_response
from newssurvey.util.sys_ import print_error, print_warning
from newssurvey.util.tiktoken_ import fit_items_to_input_token_limit

SECTION_PATTERN = re.compile(r"(?P<num>\d+)\. (?P<section>.+?)")  # Also used by refine_sections.


def are_sections_valid(numbered_sections: list[str], max_sections: int) -> bool:  # Also used by refine_sections.
    """Return true if the section names are valid, otherwise false.

    A validation error is printed if a section name is invalid.

    Example of valid section: "3. Epidemiology of daytime drowsiness"
    """
    if not numbered_sections:
        print_error("No sections found.")
        return False

    if len(numbered_sections) > max_sections:
        print_error(f"Number of sections ({len(numbered_sections)}) exceeds the maximum number of permissible sections ({max_sections}).")
        return False

    seen = set()
    for num, numbered_section in enumerate(numbered_sections, start=1):
        if not isinstance(numbered_section, str):
            print_error(f"Section {num} is not a string.")
            return False

        match = SECTION_PATTERN.fullmatch(numbered_section)
        if not match:
            print_error(f"Section line #{num} does not match expected pattern. The section string is: {numbered_section!r}")
            return False

        section_num = int(match["num"])
        if section_num != num:
            print_error(f"Section #{num} is not sequential. The section string is: {numbered_section!r}")
            return False

        section = match["section"]
        section_casefold = section.casefold()
        if section_casefold in seen:
            print_error(f"Section #{num} is a duplicate. The section string is: {numbered_section!r}")
            return False
        seen.add(section_casefold)

        if section.endswith(":"):
            print_error(f"Section #{num} ends with a colon. The section string is: {numbered_section!r}")
            return False

    return True


def _list_sections(user_query: str, source_module: ModuleType, *, titles: list[str], max_sections: int, max_attempts: int = 3) -> Optional[list[str]]:
    assert user_query
    assert titles

    prompt_data = {"user_query": user_query, "source_site_name": source_module.SOURCE_SITE_NAME, "source_type": source_module.SOURCE_TYPE}

    def prompt_formatter(titles_truncated: list[str]) -> str:
        numbered_titles = [f"{i}. {s}" for i, s in enumerate(titles_truncated, start=1)]
        numbered_titles_str = "\n".join(numbered_titles)
        prompt_data["task"] = PROMPTS["3.1. list_sections"].format(num_titles=len(titles_truncated), titles=numbered_titles_str, max_sections=max_sections)
        prompt = PROMPTS["0. common"].format(**prompt_data)
        return prompt

    model_size = "large"
    num_titles_used, prompt = fit_items_to_input_token_limit(titles, model=MODELS["text"][model_size], formatter=prompt_formatter, approach="rate")
    if num_titles_used < len(titles):
        num_titles_not_used = len(titles) - num_titles_used
        print_warning(f"Used only {num_titles_used:,}/{len(titles):,} titles for getting section names, leaving the last {num_titles_not_used:,} titles unused.")

    for num_attempt in range(1, max_attempts + 1):
        response = get_content(prompt, model_size=model_size, log=True, read_cache=(num_attempt == 1))

        if is_none_response(response):
            return

        numbered_response_sections = [line.strip() for line in response.splitlines()]
        numbered_response_sections = [line for line in numbered_response_sections if line]

        error = io.StringIO()
        with contextlib.redirect_stderr(error):
            response_is_valid = are_sections_valid(numbered_response_sections, max_sections)
        if not response_is_valid:
            error = error.getvalue().rstrip().removeprefix("Error: ")
            if num_attempt == max_attempts:
                raise LanguageModelOutputStructureError(error)
            else:
                print_warning(f"Fault in attempt {num_attempt} of {max_attempts} while getting section names: {error}")
                continue

        break

    numbered_response_matches = [SECTION_PATTERN.fullmatch(line) for line in numbered_response_sections]
    sections = [match["section"] for match in numbered_response_matches]
    return sections


def list_sections(user_query: str, source_module: ModuleType, *, titles: list[str], max_sections: int) -> list[str]:
    """Return an ordered list of section names.

    The internal functions raise `LanguageModelOutputError` if the model output has an error.
    The subclass `LanguageModelOutputRejectionError` is raised if the output is rejected.
    The subclass `LanguageModelOutputStructureError` is raised if the output is structurally invalid.
    """
    assert NUM_SECTIONS_MIN <= max_sections <= NUM_SECTIONS_MAX, (max_sections, NUM_SECTIONS_MIN, NUM_SECTIONS_MAX)

    titles = sort_by_distance(user_query, titles, model_size="large", distance="cosine")
    sections = _list_sections(user_query, source_module, titles=titles, max_sections=max_sections)

    if sections is None:
        raise LanguageModelOutputRejectionError("No sections were listed for the query based on the available articles.")

    assert sections
    return sections



from typing import Callable, Optional

import tiktoken

from newssurvey.util.openai_ import MAX_INPUT_TOKENS, MAX_OUTPUT_TOKENS

HEADER_TOKENS_PER_MESSAGE: int = 4  # Estimate as per https://platform.openai.com/docs/advanced-usage/managing-tokens.
FOOTER_TOKENS: int = 2  # Estimate as per https://platform.openai.com/docs/advanced-usage/managing-tokens.


def get_encoding(model: str) -> tiktoken.Encoding:
    """Return the encoding for the given model."""
    return tiktoken.encoding_for_model(model)


def count_tokens(text: str, *, model: str) -> int:
    """Return the number of tokens used by the given text and model."""
    encoding = get_encoding(model)
    encoded = encoding.encode(text)
    return len(encoded)


def calc_input_token_usage(text: str, *, model: str, num_output_tokens: Optional[int] = None) -> dict[str, int]:
    """Return the input token usage for the given text and model.

    `num_output_tokens` is the allocated number of tokens for the output text.
    If given, it must be less than or equal to the maximum output tokens for the given model.
    If not given, it is set to the maximum output tokens for the given model.

    The returned dictionary has the following keys:
    * `used_tokens`: Number of tokens used by input.
    * `usable_tokens`: Maximum number of usable input tokens, leaving maximum room for output tokens.
    """
    used_tokens = count_tokens(text, model=model)

    if num_output_tokens is None:
        num_output_tokens = MAX_OUTPUT_TOKENS[model]
    else:
        assert 1 <= num_output_tokens <= MAX_OUTPUT_TOKENS[model], (num_output_tokens, MAX_OUTPUT_TOKENS[model])

    assert MAX_INPUT_TOKENS[model] >= num_output_tokens

    usable_tokens = max(0, MAX_INPUT_TOKENS[model] - num_output_tokens - (HEADER_TOKENS_PER_MESSAGE * 2) - FOOTER_TOKENS)
    return {"used_tokens": used_tokens, "usable_tokens": usable_tokens}


def fit_items_to_input_token_limit(items: list, *, model: str, formatter: Callable[[list], str] = "\n".join, approach: str = "binary", num_output_tokens: Optional[int] = None) -> tuple[int, str]:
    """Return the number of items used and the text that fits the input token limit for the given items and model.

    `formatter` must be a function that takes a list of items and returns a string.
    `approach` must be either "binary" or "rate". It is the optimization approach used to fit the items to the input token limit.

    `num_output_tokens` is the allocated number of tokens for the output text.
    If given, it must be less than or equal to the maximum output tokens for the given model.
    If not given, it is set to the maximum output tokens for the given model.
    """
    # Tests:
    # _=fit_items_to_input_token_limit([string.printable]*10_000, model="gpt-4o-2024-08-06", approach='binary') -> Using 3,487/10,000 items of text for model gpt-4o-2024-08-06 and encoding o200k_base, with 111,583/111,606 tokens.
    # _=fit_items_to_input_token_limit(''.join(random.Random(0).choices(string.printable, k=1_000_000)).split('\n'), model="gpt-4o-2024-08-06", approach='binary') -> Using 1,480/9,929 items of text for model gpt-4o-2024-08-06 and encoding o200k_base, with 111,410/111,606 tokens.
    fn_calc_usage = lambda text: calc_input_token_usage(text, model=model, num_output_tokens=num_output_tokens)
    encoding = get_encoding(model).name
    text = formatter(items)
    usage = fn_calc_usage(text)
    num_items = len(items)
    if usage["used_tokens"] <= usage["usable_tokens"]:
        print(f"Using all {num_items:,} items of text for model {model} and encoding {encoding}, with {usage['used_tokens']:,}/{usage['usable_tokens']:,} tokens.")
        return num_items, text

    iteration = 0
    match approach:
        case "binary":
            # Binary search
            lo, hi = 1, num_items
            while lo < hi:
                iteration += 1
                mid = (lo + hi + 1) // 2
                mid_text = formatter(items[:mid])
                usage = fn_calc_usage(mid_text)
                num_items_used = mid
                if usage["used_tokens"] <= usage["usable_tokens"]:
                    lo = mid
                else:
                    hi = mid - 1
                print(f"Tried {num_items_used:,}/{num_items:,} items of text for model {model} in iteration {iteration:,} using {usage['used_tokens']:,}/{usage['usable_tokens']:,} tokens.")
            num_items_used = lo
        case "rate":
            # Rate-based search
            num_items_used = num_items
            items_to_excess_tokens: dict[int, int] = {}
            while True:
                if (num_items_used in items_to_excess_tokens) and (items_to_excess_tokens[num_items_used] <= 0):
                    num_items_used = max({p: e for p, e in items_to_excess_tokens.items() if e < 0}, key=items_to_excess_tokens.__getitem__)
                    candidate_num_items_used = num_items_used + 1
                    if (candidate_num_items_used in items_to_excess_tokens) and (items_to_excess_tokens[candidate_num_items_used] > 0):
                        break
                    num_items_used = candidate_num_items_used
                iteration += 1
                usage = fn_calc_usage(formatter(items[:num_items_used]))
                items_to_excess_tokens[num_items_used] = usage["used_tokens"] - usage["usable_tokens"]
                rate = usage["used_tokens"] / usage["usable_tokens"]
                print(f"Tried {num_items_used:,}/{num_items:,} items of text for model {model} in iteration {iteration:,} using {usage['used_tokens']:,}/{usage['usable_tokens']:,} tokens.")
                num_items_used = num_items_used / rate
                num_items_used = int(num_items_used)  # Note: Always using `round` or `math.ceil` instead of `int` was observed to lead to an infinite loop.
                num_items_used = min(num_items, num_items_used)
        case _:
            raise ValueError(f"Unsupported approach {approach!r}.")

    encoding = get_encoding(model).name
    text_used = formatter(items[:num_items_used])
    usage = fn_calc_usage(text_used)
    print(f"Using {num_items_used:,}/{num_items:,} items of text for model {model} and encoding {encoding}, with {usage['used_tokens']:,}/{usage['usable_tokens']:,} tokens.")
    return num_items_used, text_used



import dotenv


def load_dotenv():
    """Load the .env file normally and also from the current directory."""
    # Ref: https://stackoverflow.com/a/78972639/
    dotenv.load_dotenv()
    dotenv.load_dotenv(dotenv.find_dotenv(usecwd=True))



from typing import Optional

from newssurvey.config import NUM_SECTIONS_DEFAULT, NUM_SECTIONS_MIN, NUM_SECTIONS_MAX, OUTPUT_FORMAT_DEFAULT
from newssurvey.exceptions import InputError
from newssurvey.types import AnalyzedArticleGen1, SearchResult, SearchArticle, SectionGen1, Response
from newssurvey.util.input import get_confirmation
from newssurvey.util.openai_ import ensure_openai_key, MODELS
from newssurvey.workflow.user.query import ensure_query_is_valid
from newssurvey.workflow.user.source import ensure_source_is_valid, get_source_module
from newssurvey.workflow.user.output import format_text_output, format_output, SUPPORTED_OUTPUT_FORMATS
from newssurvey.workflow.llm.accumulate_search_terms import accumulate_search_terms
from newssurvey.workflow.llm.list_search_terms import list_search_terms
from newssurvey.workflow.llm.filter_search_results import filter_search_results
from newssurvey.workflow.llm.list_sections import list_sections
from newssurvey.workflow.llm.refine_sections import refine_sections
from newssurvey.workflow.llm.create_title import create_title
from newssurvey.workflow.llm.rate_articles import rate_articles
from newssurvey.workflow.llm.condense_articles import condense_articles
from newssurvey.workflow.llm.combine_articles import combine_articles
from newssurvey.workflow.source.get_articles import get_articles
from newssurvey.workflow.source.map_citations import map_citations


def generate_response(source: str, query: str, max_sections: int = NUM_SECTIONS_DEFAULT, output_format: Optional[str] = OUTPUT_FORMAT_DEFAULT, confirm: bool = False) -> Response:
    """Return a response for the given source and query.

    The returned response contains the attributes: format, title, response.

    The progress is printed to stdout.

    Params:
    * `source`: Name of supported news source.
    * `query`: Question or concern answerable by the news source.
    * `max_sections`: Maximum number of sections to include in the response, between 10 and 100. Its recommended value, also the default, is 100.
    * `output_format`: Output format. It can be txt (for text), md (for markdown), gfm.md (for GitHub Flavored markdown), html, pdf, or json. Its default is txt.
    * `confirm`: Confirm as the workflow progresses. If true, a confirmation is interactively sought as each step of the workflow progresses. Its default is false.

    If failed, a subclass of the `newssurvey.exceptions.Error` exception is raised.
    """
    ensure_openai_key()

    ensure_source_is_valid(source)
    print(f"SOURCE: {source}")
    source_module = get_source_module(source)

    ensure_query_is_valid(query)
    query_sep = "\n" if (len(query.splitlines()) > 1) else " "
    print(f"QUERY:{query_sep}{query}")

    if not (NUM_SECTIONS_MIN <= max_sections <= NUM_SECTIONS_MAX):
        raise InputError(f"Invalid number of sections: {max_sections}. It must be between {NUM_SECTIONS_MIN} and {NUM_SECTIONS_MAX}.")
    print(f"MAX SECTIONS: {max_sections}")

    if output_format not in SUPPORTED_OUTPUT_FORMATS:
        raise InputError(f"Invalid output format: {output_format}. It must be one of: {', '.join(SUPPORTED_OUTPUT_FORMATS)}")
    print(f"FORMAT: {output_format}")

    print(f"MODELS: text:large={MODELS["text"]["large"]}, text:small={MODELS["text"]["small"]}, embedding:large={MODELS["embedding"]["large"]}")

    search_terms: list[str] = list_search_terms(user_query=query, source_module=source_module)
    search_terms: list[str] = accumulate_search_terms(user_query=query, source_module=source_module, search_terms=search_terms)
    print(f"SEARCH TERMS ({len(search_terms)}): " + ", ".join(search_terms))

    if confirm:
        get_confirmation("getting search results")
    search_results: list[SearchResult] = filter_search_results(user_query=query, source_module=source_module, search_terms=search_terms)
    print(f"SEARCH RESULTS ({len(search_results)}):\n" + "\n".join(f'#{num}: {r['title']}\n{r['link']}\n{r['description']}' for num, r in enumerate(search_results, start=1)))

    if confirm:
        get_confirmation("getting articles")
    articles: list[SearchArticle] = get_articles(source_module=source_module, search_results=search_results)

    if confirm:
        get_confirmation("listing sections")
    sections: list[str] = list_sections(user_query=query, source_module=source_module, titles=[r["title"] for r in articles], max_sections=max_sections)
    sections: list[str] = refine_sections(user_query=query, source_module=source_module, sections=sections, titles=[r["title"] for r in articles], max_sections=max_sections)
    num_sections = len(sections)
    section_names_str = f"SECTIONS ({num_sections}):\n" + "\n".join([f"{num}: {section}" for num, section in enumerate(sections, start=1)])
    print(section_names_str)

    if confirm:
        get_confirmation("rating articles")
    articles_and_sections: list[AnalyzedArticleGen1] = rate_articles(user_query=query, source_module=source_module, articles=articles, sections=sections)
    print(f"RATED ARTICLES x SECTIONS PAIRS SUMMARY: {len(articles_and_sections)} articles x {num_sections} sections = {sum(len(a['sections']) for a in articles_and_sections):,} actual pairs / {len(articles_and_sections) * num_sections:,} possible pairs")
    print(section_names_str)

    if confirm:
        get_confirmation("condensing articles")
    articles_and_sections: list[AnalyzedArticleGen1] = condense_articles(user_query=query, source_module=source_module, articles=articles_and_sections, sections=sections)
    print("RATED SECTIONS BY ARTICLE:\n" + "\n".join(f'#{a_num}: {a["article"]["title"]} ({len(a["sections"])}/{num_sections} sections) (r={sum(s['rating'] for s in a['sections'])})\n\t{"\n\t".join(f'{s_num}. {s["section"]} (r={s["rating"]})' for s_num, s in enumerate(a["sections"], start=1))}' for a_num, a in enumerate(articles_and_sections, start=1)))
    print(f"ARTICLES BY SECTION ({num_sections}):")
    for section_num, section in enumerate(sections, start=1):
        section_articles = [a for a in articles_and_sections if any(section == s["section"] for s in a["sections"])]
        section_rating = sum(s["rating"] for a in section_articles for s in a["sections"] if section == s["section"])
        print(f"{section_num}. {section} ({len(section_articles)} articles) (r={section_rating:,})")
        section_articles.sort(key=lambda a: (next(s["rating"] for s in a["sections"] if section == s["section"]), sum(s["rating"] for s in a["sections"])), reverse=True)
        for article_num, article in enumerate(section_articles, start=1):
            article_section_pair_rating = next(s["rating"] for s in article["sections"] if section == s["section"])
            article_rating = sum(s["rating"] for s in article["sections"])
            print(f"\t{article_num}: {article['article']['title']} (r={article_section_pair_rating}/{article_rating})")
    print(f"CONDENSED ARTICLES x SECTIONS PAIRS SUMMARY: {len(articles_and_sections)} articles x {num_sections} sections = {sum(len(a['sections']) for a in articles_and_sections):,} actual pairs / {len(articles_and_sections) * num_sections:,} possible pairs")
    print(section_names_str)

    if confirm:
        get_confirmation("generating section texts")
    section_texts: list[SectionGen1] = combine_articles(user_query=query, source_module=source_module, articles=articles_and_sections, sections=sections)
    # response_text = f"{title}\n\n" + "Sections:\n" + "\n".join([f"{num}: {section}" for num, section in enumerate(sections, start=1)]) + "\n\n" + "\n\n".join(f'Section {num}. {s["title"]}:\n\n{s["text"]}' for num, s in enumerate(section_texts, start=1))
    # print(f"REPORT:\n\n{response_text}")
    section_texts, citations = map_citations(sections=section_texts)

    if confirm:
        get_confirmation("generating title")
    # Note: The title is created only after the sections are generated to ensure that the title is based on the actual remaining sections.
    title: str = create_title(user_query=query, source_module=source_module, sections=[s["title"] for s in section_texts])
    print(f"TITLE: {title}")

    response_text: str = format_text_output(title=title, sections=section_texts, citations=citations)
    print(f"REPORT:\n\n{response_text}")

    if output_format == "txt":
        response_data = response_text
    elif output_format in SUPPORTED_OUTPUT_FORMATS:
        response_data: str | bytes = format_output(title=title, sections=section_texts, citations=citations, output_format=output_format)
    else:
        raise ValueError(f"Unsupported output format: {output_format!r}")

    response: Response = Response(format=output_format, title=title, response=response_data)
    return response



import contextlib
import io
from types import ModuleType

from newssurvey.exceptions import LanguageModelOutputStructureError
from newssurvey.config import PROMPTS
from newssurvey.util.openai_ import get_content
from newssurvey.util.str import is_none_response
from newssurvey.util.sys_ import print_warning
from newssurvey.workflow.llm.list_search_terms import is_search_terms_list_valid


def _accumulate_search_terms(user_query: str, source_module: ModuleType, search_terms: list[str], max_attempts: int = 3) -> list[str]:
    """Return the accumulated search terms."""
    assert search_terms

    prompt_data = {"user_query": user_query, "source_site_name": source_module.SOURCE_SITE_NAME, "source_type": source_module.SOURCE_TYPE}
    prompt_data["task"] = PROMPTS["1.2. accumulate_search_terms"].format(**prompt_data, num_terms=len(search_terms), terms="\n".join(search_terms))
    prompt = PROMPTS["0. common"].format(**prompt_data)

    for num_attempt in range(1, max_attempts + 1):
        response = get_content(prompt, model_size="large", log=True, read_cache=(num_attempt == 1))

        if is_none_response(response):
            return []

        new_terms = [t.strip() for t in response.splitlines()]
        new_terms = [t for t in new_terms if t]
        new_terms = [t for t in new_terms if not is_none_response(t)]
        new_terms = [t for t in new_terms if t not in search_terms]

        if not new_terms:
            return []

        error = io.StringIO()
        with contextlib.redirect_stderr(error):
            response_is_valid = is_search_terms_list_valid(new_terms)
        if not response_is_valid:
            error = error.getvalue().rstrip().removeprefix("Error: ")
            if num_attempt == max_attempts:
                raise LanguageModelOutputStructureError(error)
            else:
                print_warning(f"Fault in attempt {num_attempt} of {max_attempts} while accumulating section terms: {error}")
                continue

        break

    assert new_terms
    return new_terms


def accumulate_search_terms(user_query: str, source_module: ModuleType, search_terms: list[str]) -> list[str]:
    """Return the current and accumulated search terms.

    The internal functions raise `LanguageModelOutputError` if the model output has an error.
    The subclass `LanguageModelOutputStructureError` is raised by it if the output is structurally invalid.
    """
    assert search_terms
    num_original_search_terms = len(search_terms)

    iteration = 1
    while True:
        num_old_search_terms = len(search_terms)
        # Note: It is useful for the order of search_terms to be preserved as is.
        new_search_terms = _accumulate_search_terms(user_query, source_module, search_terms)
        search_terms.extend(new_search_terms)
        print(f"Search terms counts: {iteration=} original={num_original_search_terms} old={num_old_search_terms} new={len(new_search_terms)}, total={len(search_terms)}")
        if new_search_terms:
            print(f"New search terms ({len(new_search_terms)}) in iteration {iteration}: " + ", ".join(new_search_terms))
        else:
            break
        iteration += 1

    return search_terms



import concurrent.futures
import os
import time
from typing import Literal, Optional

import openai

from newssurvey.config import CACHE_EXPIRATION_BY_TAG, CACHE_SIZES_GiB
import newssurvey.exceptions
from newssurvey.util.dict import dict_str
from newssurvey.util.diskcache_ import get_diskcache
from newssurvey.util.dotenv_ import load_dotenv
from newssurvey.util.sys_ import print_warning

load_dotenv()

ChatCompletion = openai.types.chat.chat_completion.ChatCompletion
CreateEmbeddingResponse = openai.types.create_embedding_response.CreateEmbeddingResponse

# _COLOR_LIGHT_GRAY = "\033[0;37m"
# _COLOR_LIGHT_BLUE = "\033[1;34m"
_COLOR_GRAY = "\033[0;90m"
_COLOR_RESET = "\033[0m"

_DISKCACHE = get_diskcache(__file__, size_gib=CACHE_SIZES_GiB["medium"])
MODELS = {  # Ref: https://platform.openai.com/docs/models/
    "text": {
        "deprecated": "gpt-4-0125-preview",  # Note: Can require more prompt tuning for successful use. gpt-4-turbo-2024-04-09 is worse.
        "large": "gpt-4o-2024-08-06",
        "large_previous": "gpt-4o-2024-05-13",  # For token measurement purposes.
        "small": "gpt-4o-mini-2024-07-18",
    },
    "embedding": {
        "large": "text-embedding-3-large",  # Output vector length is 3072.
        "small": "text-embedding-3-small",  # Output vector length is 1536.
    },
}
TextModelSizeType = Literal["large", "small"]
EmbeddingModelSizeType = Literal["large", "small"]

MAX_INPUT_TOKENS = {
    "gpt-4-0125-preview": 128_000,
    "gpt-4o-2024-08-06": 128_000,
    "gpt-4o-2024-05-13": 128_000,
    "gpt-4o-mini-2024-07-18": 128_000,
}
MAX_OUTPUT_TOKENS = {
    "gpt-4-0125-preview": 4096,
    "gpt-4o-2024-08-06": 16_384,
    "gpt-4o-2024-05-13": 4096,
    "gpt-4o-mini-2024-07-18": 16_384,
}
MAX_WORKERS = 16


def ensure_openai_key() -> None:
    """Raise `EnvError` if the environment variable OPENAI_API_KEY is unavailable."""
    if not os.environ.get("OPENAI_API_KEY"):
        raise newssurvey.exceptions.EnvError("The environment variable OPENAI_API_KEY is unavailable. It can optionally be defined in an .env file.")


@_DISKCACHE.memoize(expire=CACHE_EXPIRATION_BY_TAG["get_completion"], tag="get_completion")
def get_completion(prompt: str, model: str, **kwargs) -> ChatCompletion:  # Note: `model` is explicitly specified to allow model-specific caching.
    """Return the completion for the given prompt and model.

    `kwargs` are forwarded to the create call.
    ."""
    assert model in MODELS["text"].values(), model
    assert model in MAX_OUTPUT_TOKENS, model

    assert model not in kwargs
    kwargs.setdefault("max_tokens", MAX_OUTPUT_TOKENS[model])

    client = openai.OpenAI()
    print(f"Requesting completion for prompt of length {len(prompt):,} using model {model} with keyword arguments: {dict_str(kwargs)}")
    time_start = time.monotonic()
    messages = [{"role": "user", "content": prompt}]

    max_attempts = 3
    for num_attempt in range(1, max_attempts + 1):
        try:
            completion = client.chat.completions.create(model=model, messages=messages, **kwargs)  # Ref: https://platform.openai.com/docs/api-reference/chat/create
        except openai.InternalServerError as exc:
            if num_attempt < max_attempts:
                print_warning(f"Completion for prompt of length {len(prompt):,} using model {model} failed in attempt {num_attempt} of {max_attempts}: {exc}")
                time.sleep(5 * num_attempt)
                continue
            else:
                assert num_attempt == max_attempts
                raise
        else:
            break

    time_used = time.monotonic() - time_start
    print(f"Received completion for prompt of length {len(prompt):,} using model {model} in {time_used:.1f}s.")
    return completion


def get_content(prompt: str, *, model_size: TextModelSizeType, completion: Optional[ChatCompletion] = None, log: bool = False, read_cache: bool = True, **kwargs) -> str:  # Note: `model_size` is explicitly required to avoid error with an unintended model size.
    """Return the completion content for the given prompt.

    `kwargs` are forwarded to the create call.
    """
    assert model_size in MODELS["text"], model_size
    model = MODELS["text"][model_size]
    if not completion:
        cache_lookup_time_start = time.monotonic()
        cache_key = get_completion.__cache_key__(prompt, model=model, **kwargs)
        if read_cache:
            cache_key_existence_status = cache_key in _DISKCACHE
            cache_lookup_time_used = time.monotonic() - cache_lookup_time_start
            if cache_key_existence_status:
                print(f"Found cache key for prompt of length {len(prompt):,} using model {model} in {cache_lookup_time_used:.1f}s.")
            else:
                print(f"Cache key for prompt of length {len(prompt):,} using model {model} was not found in {cache_lookup_time_used:.1f}s.")
        else:
            cache_key_deletion_status = _DISKCACHE.delete(cache_key)
            cache_cleanup_time_used = time.monotonic() - cache_lookup_time_start
            if cache_key_deletion_status:
                print(f"Deleted cache key for prompt of length {len(prompt):,} using model {model} in {cache_cleanup_time_used:.1f}s.")
            else:
                print(f"Cache key for prompt of length {len(prompt):,} using model {model} was not found in {cache_cleanup_time_used:.1f}s.")
        completion = get_completion(prompt, model=model, **kwargs)
    content = completion.choices[0].message.content
    content = content.strip()
    assert content
    if log:
        print(f"\n{_COLOR_GRAY}PROMPT:\n{prompt}\nCOMPLETION:\n{content}{_COLOR_RESET}")
        # print(prefix_lines(f"PROMPT\n{prompt}\nCOMPLETION\n{content}"))
    return content


@_DISKCACHE.memoize(expire=CACHE_EXPIRATION_BY_TAG["get_embedding"], tag="get_embedding")
def get_embedding(text: str, model: str) -> CreateEmbeddingResponse:  # Note: `model` is explicitly specified to allow model-specific caching.
    """Return the embedding response for the given text."""
    assert model in MODELS["embedding"].values(), model
    client = openai.OpenAI()
    text_log = text[:100] + "..." if (len(text) > 100) else text
    print(f"Requesting embedding for text {text_log!r} of length {len(text):,} using model {model}.")
    # time_start = time.monotonic()
    response = client.embeddings.create(input=text, model=model)
    # time_used = time.monotonic() - time_start
    # print(f"Received embedding for text {text_log!r} of length {len(text):,} using model {model} in {time_used:.1f}s.")
    return response


def get_vector(text: str, *, model_size: str, embedding: Optional[CreateEmbeddingResponse] = None, log: bool = False) -> list[float]:  # Note: `model_size` is explicitly required to avoid error with an unintended model size.
    """Return the embedding vector for the given text."""
    assert model_size in MODELS["embedding"], model_size
    model = MODELS["embedding"][model_size]
    if not embedding:
        embedding = get_embedding(text, model=model)
    vector = embedding.data[0].embedding
    assert vector
    if log:
        print(f"\n{_COLOR_GRAY}TEXT:\n{text}\nEMBEDDING: {vector}{_COLOR_RESET}")
        # print(prefix_lines(f"TEXT\n{text}\nEMBEDDING: {embedding}"))
    return vector


def get_vectors_concurrently(texts: list[str], *, model_size: str, log: bool = False) -> dict[str, list[float]]:  # Note: `model_size` is explicitly required to avoid error with an unintended model size.
    """Return the embedding vectors for the given texts."""
    assert model_size in MODELS["embedding"], model_size
    fn_get_vector = lambda text: (text, get_vector(text, model_size=model_size, log=log))
    with concurrent.futures.ThreadPoolExecutor(max_workers=MAX_WORKERS) as executor:
        return dict(executor.map(fn_get_vector, texts))



import io
import contextlib
import re
from types import ModuleType

from newssurvey.exceptions import LanguageModelOutputStructureError, SourceInsufficiencyError
from newssurvey.types import SearchResult
from newssurvey.config import PROMPTS
from newssurvey.util.dict import dict_str
from newssurvey.util.openai_ import get_content
from newssurvey.util.sys_ import print_error, print_warning

_RESPONSE_PATTERN = re.compile(r"\d+(?: \d+)*")


def _is_response_valid(response: str, num_search_results: int) -> bool:
    """Return true if the response is valid, otherwise false.

    :param response: Valid example: '3 5 8'

    A validation error is printed if a search term is invalid.
    """
    if not response:
        print_error("No response exists.")
        return False

    if response != response.strip():
        print_error(f"Response is invalid because it has leading or trailing whitespace: {response!r}")
        return False

    num_response_lines = len(response.splitlines())
    if num_response_lines > 1:
        print_error(f"Response is invalid because it has multiple lines: {response!r}")
        return False

    if _RESPONSE_PATTERN.fullmatch(response) is None:
        print_error(f"Response is invalid because it does not match the expected pattern: {response!r}")
        return False

    responses = response.split(" ")
    num_responses = len(responses)
    if num_responses > num_search_results:
        print_error(f"Response is invalid because it has more entries ({num_responses}) than expected for the search results ({num_search_results}): {response!r}")
        return False

    seen = set()
    for count, response in enumerate(responses, start=1):
        assert response.isdigit()  # This is already checked by the regex.
        number = int(response)

        if number > num_search_results:
            print_error(f"Response #{count} has a value of {number} which is invalid because it is greater than the number of search results: {response!r}")
            return False

        if number in seen:
            print_error(f"Response #{count} has a value of {number} which is invalid because it is a duplicate: {responses!r}")
            return False
        seen.add(number)

        # Note: This fails sometimes when using the small gpt-4o-mini-2024-07-18 model. It is not strictly necessary.
        # if number < max(seen):
        #     print_error(f"Response {count} is invalid because it is not in ascending order: {number}")
        #     return False

    return True


def _filter_search_results(user_query: str, source_module: ModuleType, *, results: list[SearchResult], max_attempts: int = 3) -> list[SearchResult]:
    """Return the list of relevant search results.

    `LanguageModelOutputError` is raised if the model output has an error.
    The subclass `LanguageModelOutputStructureError` is raised if the output is structurally invalid.
    """
    assert user_query
    assert results
    prompt_data = {
        "user_query": user_query,
        "source_site_name": source_module.SOURCE_SITE_NAME,
        "source_type": source_module.SOURCE_TYPE,
        "search_results": "\n\n".join(f'{num}. {result['title']}\n{result['link']}\n{result.get('description', '')}'.rstrip() for num, result in enumerate(results, start=1)),  # Note: Link is included because it contains the year and month.
    }
    prompt_data["task"] = PROMPTS["2. filter_search_results"].format(**prompt_data)
    prompt = PROMPTS["0. common"].format(**prompt_data)

    for num_attempt in range(1, max_attempts + 1):
        response = get_content(prompt, model_size="small", log=(num_attempt > 1), read_cache=(num_attempt == 1))

        if response == "0":
            return []

        error = io.StringIO()
        with contextlib.redirect_stderr(error):
            response_is_valid = _is_response_valid(response, len(results))
        if not response_is_valid:
            error = error.getvalue().rstrip().removeprefix("Error: ")
            if num_attempt == max_attempts:
                raise LanguageModelOutputStructureError(error)
            else:
                print_warning(f"Fault in attempt {num_attempt} of {max_attempts} while filtering search results: {error}")
                continue

        break

    responses = response.split(" ")
    responses = [int(r) for r in responses]
    responses = [r for r in responses if (r != 0)]
    responses.sort()
    filtered_results = [results[r - 1] for r in responses]
    return filtered_results


def _filter_search_results_for_search_term(user_query: str, source_module: ModuleType, search_term: str) -> list[SearchResult]:
    results = {}

    def insert_paged_results(**kwargs) -> None:
        assert "query" not in kwargs, kwargs
        assert "page_num" not in kwargs, kwargs
        num_total_results = 0
        page_num = 1
        while True:
            page_results = source_module.get_search_results(query=search_term, page_num=page_num, **kwargs)
            if not page_results:
                break
            num_total_results += len(page_results)
            filtered_page_results = _filter_search_results(user_query=user_query, source_module=source_module, results=page_results)
            print(f"Limited {len(page_results)} original results to {len(filtered_page_results)} filtered results for page {page_num} of search term {search_term!r} with arguments: {dict_str(kwargs)}")
            if not filtered_page_results:
                break
            for result in filtered_page_results:
                result_link = result["link"]
                if result_link not in results:
                    results[result_link] = result
            page_num += 1
        print(f"Obtained {len(results)} filtered results out of {num_total_results} total results for search term {search_term!r}.")

    source_module.run_searches(insert_paged_results)

    results = list(results.values())
    return results


def filter_search_results(user_query: str, source_module: ModuleType, search_terms: list[str]) -> list[SearchResult]:
    """Return filtered search results.

    Results are filtered for relevance by the LLM.

    Returns:
        list[searchResult]: A list of dictionaries where each dictionary represents an article with its title, link, and description.

    The internal function `_filter_search_results` raises `LanguageModelOutputError` if the model output has an error.
    The subclass `LanguageModelOutputStructureError` is raised by it if the output is structurally invalid.

    `SourceInsufficiencyError` is raised if no filtered results are available.
    """
    results = {}
    num_terms = len(search_terms)
    for term_num, term in enumerate(search_terms, start=1):
        results_for_term = _filter_search_results_for_search_term(user_query=user_query, source_module=source_module, search_term=term)
        for result in results_for_term:
            result_link = result["link"]
            if result_link not in results:
                results[result_link] = result
        print(f"Accumulated a running total of {len(results)} filtered results for {term_num}/{num_terms} search terms.")

    if not results:
        raise SourceInsufficiencyError("No filtered search results exist for query.")

    results = list(results.values())
    assert len(results) == len(set(r["link"] for r in results))  # Ensures no duplicates.
    return results



import contextlib
import io

import click

import newssurvey.exceptions
from newssurvey.util.sys_ import print_error


def is_query_valid(query: str) -> bool:
    """Return true if the query is structurally valid, otherwise false.

    A validation error is printed if the query is invalid.
    """
    if not isinstance(query, str):  # Note: This happens if `-t` flag is provided without any value.
        print_error("Query must be a string.")
        return False
    if query != query.strip():
        print_error("Query must not have leading or trailing whitespace.")  # It is assumed that whitespace is stripped before this check.
        return False
    if len(query) == 0:
        print_error("No query was provided.")
        return False
    if len(query) < 2:
        print_error("Query must be at least two characters long.")
        return False
    if any(ln.lstrip().startswith("#") for ln in query.splitlines()):
        print_error("Query may not contain any line starting with # as it will interfere with separating sections in the prompts.")
        # Note:
        # - Technically, only level 1 to 2 headings need to be avoided, with level 3 and lower levels being acceptable.
        # - More technically, user supplied level 1 or 2 headings can if necessary be dedented to level 3 and lower headings.
        return False
    return True


def ensure_query_is_valid(query: str) -> None:
    """Raise `InputError` if the query is structurally invalid."""
    error = io.StringIO()
    with contextlib.redirect_stderr(error):
        if not is_query_valid(query):
            error = error.getvalue().rstrip().removeprefix("Error: ")
            raise newssurvey.exceptions.InputError(error)


def get_query(*, source_type: str, approach: str = "click.edit") -> str:
    """Get user query from user input.

    `source_type` is a string that corresponds to the SOURCE_TYPE variable of the corresponding source.

    `InputError` is raised if the approach is "click.edit" and the user-supplied query is empty.
    """
    query = None
    while not query:
        match approach:
            case "input":
                query = input(f"Specify the {source_type} question or concern: ")
            case "click.edit":
                query = click.edit(text=f"\n# Specify the {source_type} question or concern in one or more lines.\n# Lines starting with # will be skipped.", require_save=False) or ""
                query = "\n".join(ln for ln in query.splitlines() if not ln.lstrip().startswith("#"))
                if not query.strip():
                    raise newssurvey.exceptions.InputError("No query was provided.")  # Note: This is the only way for the user to abort the editor.
            case _:
                assert False, approach
        query = query.strip()
        if not is_query_valid(query):
            query = None
    return query



def dict_str(data: dict, *, sep: str = ", ") -> str:
    return sep.join(f"{k}={v}" for k, v in data.items())


def dereference_dict(input_dict: dict[str, str]) -> dict[str, str]:
    """Return a resolved version of the input dictionary, where each key is associated with its terminal value.

    In the input dictionary, values may be other keys.
    This function resolves the dictionary by following the chain of keys until it reaches a terminal value.
    If a key points to itself or is involved in a cycle, it is omitted from the output.

    Parameters:
    - input_dict (dict[str, str]): A dictionary where the keys and values are strings. The values can either be terminal values or other keys in the dictionary.

    Returns:
    - dict[str, str]: A dictionary where each key is associated with its resolved terminal value. Keys involved in cycles are not included.

    Example:
    >>> dereference_dict({'a': 'b', 'b': 'c', 'c': 'a',    'd': 'e', 'e': 'f', 'f': 'g',    'h': 'i', 'i': 'h',    'j': 'j',    'k': 'l',     'm': 'n', 'n': 'o', 'o': 'n',    'p': 'q', 'q': 'r', 'r': 's', 's': 'q'})
    {'d': 'g', 'e': 'g', 'f': 'g', 'k': 'l'}
    """
    output_dict = {}
    for start_key, start_value in input_dict.items():
        current_key, current_value = start_key, start_value
        visited = set()

        while True:
            if current_key in visited:
                break
            visited.add(current_key)

            if current_value in input_dict:
                current_key, current_value = current_value, input_dict[current_value]
            else:
                output_dict[start_key] = current_value
                break

    return output_dict



import re

from newssurvey.config import CITATION_OPEN_CHAR, CITATION_CLOSE_CHAR, CITATION_GROUP_PATTERN
from newssurvey.types import CitationGen2, SectionGen1, SectionGen2


def map_citations(sections: list[SectionGen1]) -> tuple[list[SectionGen2], list[CitationGen2]]:
    """Return transformed and mapped citations used in the individual sections to a globally consistent list of citations."""
    link_citation_map: dict[str, CitationGen2] = {}

    num_sections = len(sections)
    sections_old = sections
    del sections
    sections_new = []

    for section_num, section_old in enumerate(sections_old, start=1):
        citation_nums_old = [citation_num for citation_group in CITATION_GROUP_PATTERN.findall(section_old["text"]) for citation_num in map(int, citation_group.split(","))]
        citation_nums_old = list(dict.fromkeys(citation_nums_old))  # Remove duplicates while preserving order. Order preservation is important, otherwise the new citation numbers will be out of order.
        citation_nums_new = []
        for citation_num_old in citation_nums_old:
            assert citation_num_old > 0
            citation_old = section_old["citations"][citation_num_old - 1]
            citation_link = citation_old["link"]
            if citation_link in link_citation_map:
                citation_num_new = link_citation_map[citation_link]["number"]
                citation_map_type = "existing"
            else:
                citation_num_new = len(link_citation_map) + 1
                link_citation_map[citation_link] = CitationGen2(number=citation_num_new, title=citation_old["title"], link=citation_link)
                citation_map_type = "new"
            print(f'Mapped citation {citation_num_old} in section {section_num} {section_old["title"]!r} to {citation_map_type} citation {citation_num_new}.')
            assert citation_num_new > 0
            citation_nums_new.append(citation_num_new)

        assert len(citation_nums_old) == len(citation_nums_new)
        citation_nums_map = dict(zip(citation_nums_old, citation_nums_new))

        def repl(match: re.Match) -> str:
            """Replace the old citation numbers in the citation group with their corresponding new citation numbers."""
            # Note: To output map of old→new, add prefix: str(int(citation_num)) + '→' +
            return CITATION_OPEN_CHAR + ",".join(str(citation_nums_map[int(citation_num)]) for citation_num in match.group(1).split(",")) + CITATION_CLOSE_CHAR

        new_text = CITATION_GROUP_PATTERN.sub(repl, section_old["text"])
        section_new = SectionGen2(title=section_old["title"], text=new_text)
        sections_new.append(section_new)
        print(f'Mapped {len(citation_nums_old)} citations for section {section_num}/{num_sections} {section_old["title"]!r}.')

    assert len(sections_new) == len(sections_old)
    citations = list(link_citation_map.values())
    assert all(citation["number"] == num for num, citation in enumerate(citations, start=1))

    return sections_new, citations



import contextlib
import concurrent.futures
import itertools
import io
import re
from types import ModuleType

from newssurvey.config import PROMPTS
from newssurvey.exceptions import LanguageModelOutputStructureError, SourceInsufficiencyError
from newssurvey.types import SearchArticle, AnalyzedArticleGen1, AnalyzedSectionGen1
from newssurvey.util.openai_ import get_content, MAX_WORKERS
from newssurvey.util.sys_ import print_error, print_warning

_INPUT_SECTION_PATTERN = re.compile(r"(?P<num>\d+)\. (?P<section>.+?)")
_OUTPUT_SECTION_PATTERN = re.compile(r"(?P<num>\d+)\. (?P<section>.+?) → (?P<rating>\d{1,3})")


def _are_sections_valid(numbered_input_sections: list[str], numbered_output_sections: list[str]) -> bool:
    """Return true if the output sections have a valid rating of the input sections, otherwise false.

    A validation error is printed if a rating is invalid.
    """
    if not numbered_input_sections:
        print_error("No input section names exist.")
        return False

    if not numbered_output_sections:
        print_error("No output section names exist.")
        return False

    num_input_sections, num_output_sections = len(numbered_input_sections), len(numbered_output_sections)
    if num_input_sections != num_output_sections:
        print_error(f"The number of input sections ({num_input_sections}) and output sections ({num_output_sections}) are unequal.")
        return False

    for num, (numbered_input_section, numbered_output_section) in enumerate(itertools.zip_longest(numbered_input_sections, numbered_output_sections), start=1):
        if not isinstance(numbered_input_section, str):  # Can be None if the list is exhausted given the use of zip_longest.
            print_error(f"Input section #{num} is missing.")
            return False
        if not isinstance(numbered_output_section, str):  # Can be None if the list is exhausted given the use of zip_longest.
            print_error(f"Output section #{num} is missing.")
            return False

        input_match = _INPUT_SECTION_PATTERN.fullmatch(numbered_input_section)
        if not input_match:
            print_error(f"Input section string #{num} is invalid: {numbered_input_section!r}")
            return False
        output_match = _OUTPUT_SECTION_PATTERN.fullmatch(numbered_output_section)
        if not output_match:
            print_error(f"Output section string #{num} is invalid: {numbered_output_section!r}")
            return False

        input_num, output_num = int(input_match.group("num")), int(output_match.group("num"))
        if num != input_num:
            print_error(f"The expected input section number ({num}) and actual input section number ({input_num}) are unequal. The input section string is: {numbered_input_section!r}")
            return False
        if input_num != output_num:
            print_error(f"The input section number ({input_num}) and output section number ({output_num}) are unequal. The output section string is: {numbered_output_section!r}")
            return False

        input_section = input_match.group("section")
        if input_section != input_section.strip():
            print_error(f"The #{num} input section name has leading or trailing whitespace: {input_section!r}")
            return False
        if not input_section:
            print_error(f"The #{num} input section name is empty. The input section string is: {numbered_input_section!r}")
            return False
        output_section = output_match.group("section")
        if input_section != output_section:
            print_error(f"The #{num} input section name ({input_section!r}) and output section name ({output_section!r}) are unequal. The output section string is: {numbered_output_section!r}")
            return False

        rating = int(output_match.group("rating"))
        if not (0 <= rating <= 100):
            print_error(f"Section #{num} has an invalid output rating of {rating}. The output section string is: {numbered_output_section!r}")
            return False

    return True


def _rate_article(user_query: str, source_module: ModuleType, article: SearchArticle, sections: list[str], *, max_attempts: int = 3) -> list[dict[str, int]]:
    assert user_query
    assert sections

    numbered_input_sections = [f"{i}. {s}" for i, s in enumerate(sections, start=1)]
    numbered_input_sections_str = "\n".join(numbered_input_sections)

    prompt_data = {"user_query": user_query, "source_site_name": source_module.SOURCE_SITE_NAME, "source_type": source_module.SOURCE_TYPE}
    assert article["text"].startswith(article["title"]), article  # If this fails, fix the parsing to ensure it is true.

    prompt_data["task"] = PROMPTS["4. rate_article"].format(num_sections=len(sections), sections=numbered_input_sections_str, article=article["text"])
    prompt = PROMPTS["0. common"].format(**prompt_data)

    for num_attempt in range(1, max_attempts + 1):
        response = get_content(prompt, model_size="small", log=(num_attempt > 1), read_cache=(num_attempt == 1))
        numbered_output_sections = [line.strip() for line in response.splitlines()]
        numbered_output_sections = [line for line in numbered_output_sections if line]

        error = io.StringIO()
        with contextlib.redirect_stderr(error):
            response_is_valid = _are_sections_valid(numbered_input_sections, numbered_output_sections)
        if not response_is_valid:
            error = error.getvalue().rstrip().removeprefix("Error: ")
            if num_attempt == max_attempts:
                raise LanguageModelOutputStructureError(error)
            else:
                print_warning(f"Fault in attempt {num_attempt} of {max_attempts} while getting section ratings: {error}")
                continue

        break

    output_matches = [_OUTPUT_SECTION_PATTERN.fullmatch(line) for line in numbered_output_sections]
    rated_sections = [AnalyzedSectionGen1(section=match.group("section"), rating=int(match.group("rating"))) for match in output_matches]

    assert len(rated_sections) == len(sections)
    assert [s["section"] for s in rated_sections] == sections
    return rated_sections


def rate_articles(user_query: str, source_module: ModuleType, *, articles: list[SearchArticle], sections: list[str]) -> list[AnalyzedArticleGen1]:
    """Return a list of dictionaries containing the search article and rated final section names.

    The rating represents how well the article can contribute to the section in the context of the user query.

    Articles with no positively rated sections are skipped.

    The internal functions raise `LanguageModelOutputError` if the model output has an error.
    Specifically, its subclass `LanguageModelOutputStructureError` is raised by it if the output is structurally invalid.

    `SourceInsufficiencyError` is raised if no usable articles remain for the query.
    """
    # articles = sorted(articles, key=lambda a: len(a["link"]), reverse=True)  # For reproducible testing, but not necessary for cache.
    num_articles, num_sections = len(articles), len(sections)
    rated_articles: list[AnalyzedArticleGen1] = []
    skipped_articles: list[SearchArticle] = []

    def rate_article(article: SearchArticle) -> AnalyzedArticleGen1:
        rated_sections = _rate_article(user_query=user_query, source_module=source_module, article=article, sections=sections)
        rated_sections = [s for s in rated_sections if s["rating"] > 0]
        return AnalyzedArticleGen1(article=article, sections=rated_sections)

    with concurrent.futures.ThreadPoolExecutor(max_workers=MAX_WORKERS) as executor:
        futures = {executor.submit(rate_article, article): article for article in articles}
        for future_num, future in enumerate(concurrent.futures.as_completed(futures), start=1):
            analyzed_article = future.result()
            if rated_sections := analyzed_article["sections"]:
                print(f"#{future_num}/{num_articles}: {analyzed_article['article']['title']} ({len(rated_sections)}/{num_sections} sections):\n\t" + "\n\t".join(f'{section_num}. {s['section']} (r={s['rating']})' for section_num, s in enumerate(rated_sections, start=1)))
                rated_articles.append(analyzed_article)
            else:
                print(f"No positively rated section names exist for article #{future_num}/{num_articles}: {analyzed_article['article']['title']}")
                skipped_articles.append(analyzed_article["article"])

    print(f"{len(skipped_articles)}/{num_articles} articles were skipped due to no positively rated sections:\n\t" + "\n\t".join(f'{num}. {a["title"]}' for num, a in enumerate(skipped_articles, start=1)))
    print(f"{len(rated_articles)}/{num_articles} articles remain with positively rated sections.")

    rated_articles.sort(key=lambda a: sum(s["rating"] for s in a["sections"]), reverse=True)

    if not rated_articles:
        raise SourceInsufficiencyError("No usable articles remain for the query.")
    return rated_articles



def triangular_number(n: int) -> int:
    """Return the nth triangular number.

    The nth triangular number is the sum of the first n positive integers.

    Example: triangular_number(3) == 6
    """
    return n * (n + 1) // 2



import itertools


def get_batches(iterable, batch_size: int, include_incomplete: bool = True) -> list[list]:
    batches = [list(batch) for batch in itertools.batched(iterable, batch_size)]  # Note: Using `list` avoids returning tuples.
    if (not include_incomplete) and batches and (len(batches[-1]) < batch_size):
        batches.pop()
    return batches



import dataclasses
from typing import NotRequired, Required, TypedDict


class SearchResult(TypedDict):
    title: Required[str]
    link: Required[str]
    description: NotRequired[str]


class SearchArticle(SearchResult):
    text: Required[str]


class AnalyzedSectionGen1(TypedDict):
    section: Required[str]
    rating: Required[int]


class AnalyzedArticleGen1(TypedDict):
    article: Required[SearchArticle]
    sections: Required[list[AnalyzedSectionGen1]]


class AnalyzedSectionGen2(TypedDict):
    section: Required[str]
    rating: Required[int]
    text: Required[str]


class AnalyzedArticleGen2(TypedDict):
    article: Required[SearchArticle]
    sections: Required[list[AnalyzedSectionGen2]]


class CitationGen1(TypedDict):
    title: Required[str]
    link: Required[str]


class SectionGen1(TypedDict):
    title: Required[str]
    text: Required[str]
    citations: Required[list[CitationGen1]]


class CitationGen2(TypedDict):
    number: Required[int]
    title: Required[str]
    link: Required[str]


class SectionGen2(TypedDict):
    title: Required[str]
    text: Required[str]


@dataclasses.dataclass(frozen=True, kw_only=True)
class Response:
    format: str
    title: str  # Included to facilitate writing the response to file with the title in its name.
    response: str | bytes



import contextlib
import io
from types import ModuleType

from newssurvey.config import PROMPTS, NUM_SECTIONS_MIN, NUM_SECTIONS_MAX
from newssurvey.exceptions import LanguageModelOutputStructureError, LanguageModelOutputConvergenceError
from newssurvey.util.difflib_ import ndiffstr
from newssurvey.util.openai_ import get_content, MODELS
from newssurvey.util.scipy_ import sort_by_distance
from newssurvey.util.sys_ import print_warning
from newssurvey.util.textwrap import tab_indent
from newssurvey.util.tiktoken_ import fit_items_to_input_token_limit
from newssurvey.workflow.llm.list_sections import SECTION_PATTERN, are_sections_valid


def _refine_sections(user_query: str, source_module: ModuleType, *, sections: list[str], titles: list[str], max_sections: int, max_attempts: int = 3) -> list[str]:
    assert user_query
    assert sections
    assert titles

    prompt_data = {"user_query": user_query, "source_site_name": source_module.SOURCE_SITE_NAME, "source_type": source_module.SOURCE_TYPE}
    numbered_sections = [f"{i}. {s}" for i, s in enumerate(sections, start=1)]
    numbered_sections_str = "\n".join(numbered_sections)

    def prompt_formatter(titles_truncated: list[str]) -> str:
        numbered_titles = [f"{i}. {s}" for i, s in enumerate(titles_truncated, start=1)]
        numbered_titles_str = "\n".join(numbered_titles)
        prompt_data["task"] = PROMPTS["3.2. refine_sections"].format(num_sections=len(sections), sections=numbered_sections_str, num_titles=len(titles_truncated), titles=numbered_titles_str, max_sections=max_sections)
        prompt = PROMPTS["0. common"].format(**prompt_data)
        return prompt

    model_size = "large"
    num_titles_used, prompt = fit_items_to_input_token_limit(titles, model=MODELS["text"][model_size], formatter=prompt_formatter, approach="rate")
    if num_titles_used < len(titles):
        num_titles_not_used = len(titles) - num_titles_used
        print_warning(f"Used only {num_titles_used:,}/{len(titles):,} titles for getting section names, leaving the last {num_titles_not_used:,} titles unused.")

    for num_attempt in range(1, max_attempts + 1):
        response = get_content(prompt, model_size=model_size, log=True, read_cache=(num_attempt == 1))

        numbered_response_sections = [line.strip() for line in response.splitlines()]
        numbered_response_sections = [line for line in numbered_response_sections if line]

        error = io.StringIO()
        with contextlib.redirect_stderr(error):
            response_is_valid = are_sections_valid(numbered_response_sections, max_sections)
        if not response_is_valid:
            error = error.getvalue().rstrip().removeprefix("Error: ")
            if num_attempt == max_attempts:
                raise LanguageModelOutputStructureError(error)
            else:
                print_warning(f"Fault in attempt {num_attempt} of {max_attempts} while refining section names: {error}")
                continue

        break

    numbered_response_matches = [SECTION_PATTERN.fullmatch(line) for line in numbered_response_sections]
    sections = [match["section"] for match in numbered_response_matches]
    return sections


def refine_sections(user_query: str, source_module: ModuleType, *, sections: list[str], titles: list[str], max_sections: int) -> list[str]:
    """Return a refined ordered list of section names.

    The internal functions raise `LanguageModelOutputError` if the model output has an error.
    The subclass `LanguageModelOutputStructureError` is raised if the output is structurally invalid.
    """
    assert NUM_SECTIONS_MIN <= max_sections <= NUM_SECTIONS_MAX, (max_sections, NUM_SECTIONS_MIN, NUM_SECTIONS_MAX)
    assert sections
    num_original_sections = len(sections)
    assert num_original_sections <= max_sections, (num_original_sections, max_sections)  # Note: NUM_SECTIONS_MIN is intentionally not checked.

    titles = sort_by_distance(user_query, titles, model_size="large", distance="cosine")

    iteration = 1
    max_iterations = 20  # Max observed = 10
    while True:
        if iteration > max_iterations:
            raise LanguageModelOutputConvergenceError(f"Failed to converge refining section names after {max_iterations} iterations.")
        num_old_sections = len(sections)
        new_sections = _refine_sections(user_query, source_module, sections=sections, titles=titles, max_sections=max_sections)
        assert new_sections
        num_new_sections = len(new_sections)
        assert num_new_sections <= max_sections, (num_new_sections, max_sections)
        print(f"Section counts in {iteration=}/{max_iterations}: original={num_original_sections} old={num_old_sections} new={num_new_sections}")
        if sections == new_sections:
            print(f"Section changes in {iteration=}/{max_iterations}: (none)")
            break
        else:
            diff = ndiffstr(sections, new_sections)
            print(f"Section changes in {iteration=}/{max_iterations}:\n{tab_indent(diff)}")
            sections = new_sections
            iteration += 1

    assert sections
    return sections



import concurrent.futures
import contextlib
import io
import re
from types import ModuleType

from newssurvey.config import PROMPTS, CITATION_OPEN_CHAR, CITATION_CLOSE_CHAR, CITATION_GROUP_PATTERN
from newssurvey.exceptions import LanguageModelOutputLimitError, LanguageModelOutputStructureError
from newssurvey.types import AnalyzedArticleGen2, CitationGen1, SectionGen1
from newssurvey.util.openai_ import get_content, MODELS, MAX_OUTPUT_TOKENS, MAX_WORKERS
from newssurvey.util.sys_ import print_warning, print_error
from newssurvey.util.textwrap import tab_indent
from newssurvey.util.tiktoken_ import count_tokens, fit_items_to_input_token_limit

_MODEL_SIZE = [
    "small",  # Do not use. Does not generate citations well.
    "large",  # Good.
    "deprecated",  # Do not use. Does not follow instructions equally well as 4o. Does not generate citations.
][1]
_MODEL = MODELS["text"][_MODEL_SIZE]

_INVALID_BRACKETS = ["〖〗"]  # These have been observed in the output.
_INVALID_BRACKETS_PATTERNS = {f"{invalid_open_bracket}{invalid_close_bracket}": re.compile(CITATION_GROUP_PATTERN.pattern.translate(str.maketrans(f"{CITATION_OPEN_CHAR}{CITATION_CLOSE_CHAR}", f"{invalid_open_bracket}{invalid_close_bracket}"))) for invalid_open_bracket, invalid_close_bracket in _INVALID_BRACKETS}


def _is_output_valid(text: str, *, section: str, num_articles: int) -> bool:
    """Return true if the output text is valid, otherwise false.

    A validation error is printed if the output text is invalid.
    """
    if text != text.strip():
        print_error(f"The text for the section {section!r} has leading or trailing whitespace.")
        return False

    if not text:
        print_error(f"The text for the section {section!r} is empty.")
        return False

    if "\n1. **" in text:
        print_error(f"The text for the section {section!r} contains a markdown list item.")
        return False

    if text.startswith(CITATION_OPEN_CHAR):
        print_error(f"The text for the section {section!r} starts with an opening bracket meant for a citation group.")
        return False

    if text.endswith(CITATION_CLOSE_CHAR):
        print_error(f"The text for the section {section!r} ends with a closing bracket meant for a citation group.")
        return False

    if f"{CITATION_OPEN_CHAR}{CITATION_CLOSE_CHAR}" in text:
        print_error(f"The text for the section {section!r} contains an empty citation group.")
        return False

    if f"{CITATION_CLOSE_CHAR}{CITATION_OPEN_CHAR}" in text:
        print_error(f"The text for the section {section!r} could contain two adjacent citation groups.")
        return False

    num_opens, num_closes = text.count(CITATION_OPEN_CHAR), text.count(CITATION_CLOSE_CHAR)
    if num_opens != num_closes:
        print_error(f"The text for the section {section!r} has {num_opens} opening brackets but {num_closes} closing brackets meant for citation groups.")
        return False

    # Ensure citation brackets are balanced.
    # Passing example: "〚1,2,3〛"
    # Failing example: "〚1,2,〚3〛〛"
    balance = 0
    for char in text:
        if char == CITATION_OPEN_CHAR:
            balance += 1
        elif char == CITATION_CLOSE_CHAR:
            balance -= 1
        if balance not in (0, 1):
            print_error(f"The text for the section {section!r} has unbalanced citation brackets.")
            return False

    # Note: Having no citations is allowed for now.

    citation_groups = CITATION_GROUP_PATTERN.findall(text)
    for num_citation_group, citation_group_str in enumerate(citation_groups, start=1):
        if not citation_group_str:
            print_error(f"The citation group #{num_citation_group} for the section {section!r} is empty.")
            return False

        citation_group = [g.strip() for g in citation_group_str.split(",")]
        for citation_str in citation_group:
            if not citation_str.isdigit():
                print_error(f"The citation {citation_str!r} in citation group #{num_citation_group} ({citation_group_str!r}) for the section {section!r} is not a number.")
                return False

            citation = int(citation_str)
            if not 1 <= citation <= num_articles:
                print_error(f"The citation {citation} in citation group #{num_citation_group} ({citation_group_str!r}) for the section {section!r} is not within the expected range of 1 to {num_articles}.")
                return False

            # Note: Duplicates are not checked because they can be managed.

    # Check for invalid brackets
    for brackets, pattern in _INVALID_BRACKETS_PATTERNS.items():
        if pattern.search(text):
            print_error(f"The text for the section {section!r} contains invalid brackets {brackets}.")
            # Note: A regex substitution could in principle be used to replace the invalid brackets with valid ones, but it is not used so as to ensure that the LLM is paying attention.
            return False

    return True


def _combine_articles(user_query: str, source_module: ModuleType, *, sections: list[str], section: str, articles: list[str], max_attempts: int = 3) -> tuple[int, str]:
    assert user_query
    assert section

    prompt_data = {"user_query": user_query, "source_site_name": source_module.SOURCE_SITE_NAME, "source_type": source_module.SOURCE_TYPE}
    num_sections = len(sections)
    numbered_sections = [f"{num}. {s}" for num, s in enumerate(sections, start=1)]
    numbered_sections_str = "\n".join(numbered_sections)
    section_number = sections.index(section) + 1
    numbered_section = f"{section_number}. {section}"
    max_output_tokens = min(8192, MAX_OUTPUT_TOKENS[_MODEL])  # Max observed: <800 output tokens.

    def prompt_formatter(articles_truncated: list[str]) -> str:
        numbered_articles = "\n\n---\n\n".join([f"[ARTICLE {num}]\n\n{article}" for num, article in enumerate(articles_truncated, start=1)])
        prompt_data["task"] = PROMPTS["6. combine_articles"].format(max_output_tokens=max_output_tokens, num_sections=num_sections, sections=numbered_sections_str, section=numbered_section, num_articles=len(articles_truncated), articles=numbered_articles)
        prompt = PROMPTS["0. common"].format(**prompt_data)
        return prompt

    num_articles_used, prompt = fit_items_to_input_token_limit(articles, model=_MODEL, formatter=prompt_formatter, approach="rate", num_output_tokens=max_output_tokens)

    for num_attempt in range(1, max_attempts + 1):
        print(f"Generating section {section!r} from {num_articles_used} used articles out of {len(articles)} supplied articles using the {_MODEL_SIZE} model {_MODEL} in attempt {num_attempt}.")
        response = get_content(prompt, model_size=_MODEL_SIZE, max_tokens=max_output_tokens, log=(num_attempt > 1), read_cache=(num_attempt == 1))
        # Note:
        # Specifying frequency_penalty<0 produced garbage output or otherwise takes forever to return.
        # Specifying presence_penalty<0 helped produce more tokens only with presence_penalty=-2 which is risky to use.

        error = io.StringIO()
        with contextlib.redirect_stderr(error):
            response_is_valid = _is_output_valid(response, section=section, num_articles=num_articles_used)
        if not response_is_valid:
            error = error.getvalue().rstrip().removeprefix("Error: ")
            if num_attempt == max_attempts:
                raise LanguageModelOutputStructureError(error)
            else:
                print_warning(f"Fault in attempt {num_attempt} of {max_attempts} while getting a section: {error}")
                continue

        break

    num_response_tokens = count_tokens(response, model=_MODEL)
    max_safe_output_tokens_rate = 0.8
    max_safe_output_tokens = int(max_output_tokens * max_safe_output_tokens_rate)
    if num_response_tokens > max_safe_output_tokens:
        # Note: The output is intentionally not retried in this case, as `max_output_tokens` likely is insufficient. If this is reached, it may need to be checked and increased.
        raise LanguageModelOutputLimitError(f"The generated section {section!r} has {num_response_tokens:,} tokens, which is more than {max_safe_output_tokens:,} tokens which is {max_safe_output_tokens_rate:.0%} of the maximum output token limit of {max_output_tokens:,} tokens. This is unexpected.")

    return num_articles_used, response


def combine_articles(user_query: str, source_module: ModuleType, *, articles: list[AnalyzedArticleGen2], sections: list[str]) -> list[SectionGen1]:
    """Return a list of sections with texts generated from the given articles.

    The internal functions raise `LanguageModelOutputError` if the model output has an error.
    Specifically, its subclass `LanguageModelOutputLimitError` or `LanguageModelOutputStructureError` is raised by it if the output is structurally invalid.
    """
    num_articles = len(articles)
    num_sections = len(sections)
    section_texts = []

    # Add ratings to articles based on their sections
    for article in articles:
        article["article"]["rating"] = sum(article_section["rating"] for article_section in article["sections"])

    def process_section(section_num: int, section: str) -> SectionGen1:
        section_articles = []
        for article in articles:
            for article_section in article["sections"]:
                if article_section["section"] == section:
                    assert article_section["rating"] > 0
                    section_article = {"article": article["article"], "section": article_section}
                    section_articles.append(section_article)
                    break
        assert section_articles, section
        num_section_articles = len(section_articles)
        section_articles.sort(key=lambda a: (a["section"]["rating"], a["article"]["rating"], a["article"]["link"]), reverse=True)  # Link is used as a unique tiebreaker for reproducibility to facilitate a cache hit. It is also used because it often contains the article's publication date.
        section_articles_texts = [f'{a["article"]["title"]}\n\n{a["section"]["text"]}' for a in section_articles]
        num_section_articles_used, section_text = _combine_articles(user_query, source_module, sections=sections, section=section, articles=section_articles_texts)

        # Token counting and logging
        num_section_text_tokens = count_tokens(section_text, model=_MODEL)
        print(f"Generated section {section_num}/{num_sections} {section!r} from {num_section_articles_used} used articles out of {num_section_articles} supplied articles out of {num_articles} total articles, with {num_section_text_tokens:,} tokens generated using the {_MODEL_SIZE} model {_MODEL}:\n{tab_indent(section_text)}")

        # Get the articles used in this section and prepare the final section data
        section_articles_used = [a["article"] for a in section_articles[:num_section_articles_used]]
        section_citations = [CitationGen1(title=a["title"], link=a["link"]) for a in section_articles_used]
        section_data = SectionGen1(title=section, text=section_text, citations=section_citations)
        return section_data

    # Use concurrent section processing
    max_workers = min(8, MAX_WORKERS)
    with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
        futures = [executor.submit(process_section, section_num, section) for section_num, section in enumerate(sections, start=1)]
        section_texts = [future.result() for future in futures]

    return section_texts



def prefix_lines(string: str, /, *, prefix: str = "> ") -> str:
    """Return the given string with each line prefixed by the given prefix."""
    return "\n".join(prefix + line for line in string.splitlines())


def is_none_response(text: str) -> bool:
    """Return true if the given text is a "none" response, otherwise false."""
    return text.lower() in ("none", "none.")


def is_ok_response(text: str) -> bool:
    """Return true if the given text is an "OK" response, otherwise false."""
    return text.upper() in ("OK", "OK.")



import contextlib
import io
from types import ModuleType

from newssurvey.exceptions import LanguageModelOutputRejectionError, LanguageModelOutputStructureError
from newssurvey.config import PROMPTS
from newssurvey.util.openai_ import get_content
from newssurvey.util.str import is_none_response
from newssurvey.util.sys_ import print_error


def is_search_terms_list_valid(terms: list[str]) -> bool:  # Also used by accumulate_search_terms.
    """Return true if the search terms are structurally valid, otherwise false.

    A validation error is printed if a search term is invalid.
    """
    if not terms:
        print_error("No search terms exist.")
        return False

    seen = set()
    for term in terms:
        if term != term.strip():
            print_error(f"Search term is invalid because it has leading or trailing whitespace: {term!r}")
            return False

        if term.startswith(("- ", "* ")):
            print_error(f"Search term is invalid because it has a leading bullet prefix: {term}")
            return False

        if term in seen:
            print_error(f"Search term is invalid because it is a duplicate: {term}")
            return False
        seen.add(term)

    return True


def list_search_terms(user_query: str, source_module: ModuleType) -> list[str]:
    """Return the list of search terms for the user query.

    `LanguageModelOutputError` is raised if the model output has an error.
    The subclass `LanguageModelOutputRejectionError` is raised if the output is rejected.
    The subclass `LanguageModelOutputStructureError` is raised if the output is structurally invalid.
    """
    assert user_query
    prompt_data = {"user_query": user_query, "source_site_name": source_module.SOURCE_SITE_NAME, "source_type": source_module.SOURCE_TYPE}
    prompt_data["task"] = PROMPTS["1.1. list_search_terms"].format(**prompt_data)
    prompt = PROMPTS["0. common"].format(**prompt_data)

    response = get_content(prompt, model_size="large", log=True)

    if is_none_response(response):
        raise LanguageModelOutputRejectionError("No search terms exist for query.")

    none_responses = ("none", "none.")
    invalid_terms = ("", *none_responses)
    terms = [t.strip() for t in response.splitlines() if t.strip().lower() not in invalid_terms]  # Note: A terminal "None" line has been observed with valid entries before it.

    error = io.StringIO()
    with contextlib.redirect_stderr(error):
        if not is_search_terms_list_valid(terms):
            error = error.getvalue().rstrip().removeprefix("Error: ")
            raise LanguageModelOutputStructureError(error)

    assert terms
    return terms
